"""
HDF5 data manager for WEST.

Original HDF5 implementation: Joseph W. Kaus
Current implementation: Matthew C. Zwier

WEST exclusively uses the cross-platform, self-describing file format HDF5
for data storage. This ensures that data is stored efficiently and portably
in a manner that is relatively straightforward for other analysis tools
(perhaps written in C/C++/Fortran) to access.

The data is laid out in HDF5 as follows:
    - summary -- overall summary data for the simulation
    - /iterations/ -- data for individual iterations, one group per iteration under /iterations
        - iter_00000001/ -- data for iteration 1
            - seg_index -- overall information about segments in the iteration, including weight
            - pcoord -- progress coordinate data organized as [seg_id][time][dimension]
            - wtg_parents -- data used to reconstruct the split/merge history of trajectories
            - recycling -- flux and event count for recycled particles, on a per-target-state basis
            - auxdata/ -- auxiliary datasets (data stored on the 'data' field of Segment objects)

The file root object has an integer attribute 'west_file_format_version' which can be used to
determine how to access data even as the file format (i.e. organization of data within HDF5 file)
evolves.

Version history:
    Version 9
        - Basis states are now saved as iter_segid instead of just segid as a pointer label.
        - Initial states are also saved in the iteration 0 file, with a negative sign.
    Version 8
        - Added external links to trajectory files in iterations/iter_* groups, if the HDF5
          framework was used.
        - Added an iter group for the iteration 0 to store conformations of basis states.
    Version 7
        - Removed bin_assignments, bin_populations, and bin_rates from iteration group.
        - Added new_segments subgroup to iteration group
    Version 6
        - ???
    Version 5
        - moved iter_* groups into a top-level iterations/ group,
        - added in-HDF5 storage for basis states, target states, and generated states
"""

import logging
import pickle
import posixpath
import sys
import threading
import time
import builtins
from operator import attrgetter
from os.path import relpath, dirname

import h5py
from h5py import h5s
import numpy as np

from . import h5io
from .segment import Segment
from .states import BasisState, TargetState, InitialState
from .we_driver import NewWeightEntry
from .propagators.executable import ExecutablePropagator

import westpa


log = logging.getLogger(__name__)

file_format_version = 9

makepath = ExecutablePropagator.makepath


class flushing_lock:
    def __init__(self, lock, fileobj):
        self.lock = lock
        self.fileobj = fileobj

    def __enter__(self):
        self.lock.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.fileobj.flush()
        self.lock.release()


class expiring_flushing_lock:
    def __init__(self, lock, flush_method, nextsync):
        self.lock = lock
        self.flush_method = flush_method
        self.nextsync = nextsync

    def __enter__(self):
        self.lock.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if time.time() > self.nextsync:
            self.flush_method()
        self.lock.release()


# Data types for use in the HDF5 file
seg_id_dtype = np.int64  # Up to 9 quintillion segments per iteration; signed so that initial states can be stored negative
n_iter_dtype = np.uint32  # Up to 4 billion iterations
weight_dtype = np.float64  # about 15 digits of precision in weights
utime_dtype = np.float64  # ("u" for Unix time) Up to ~10^300 cpu-seconds
vstr_dtype = h5py.special_dtype(vlen=str)
h5ref_dtype = h5py.special_dtype(ref=h5py.Reference)
binhash_dtype = np.dtype('|S64')

# seg_status_dtype    = h5py.special_dtype(enum=(np.uint8, Segment.statuses))
# seg_initpoint_dtype = h5py.special_dtype(enum=(np.uint8, Segment.initpoint_types))
# seg_endpoint_dtype  = h5py.special_dtype(enum=(np.uint8, Segment.endpoint_types))
# istate_type_dtype   = h5py.special_dtype(enum=(np.uint8, InitialState.istate_types))
# istate_status_dtype = h5py.special_dtype(enum=(np.uint8, InitialState.istate_statuses))

seg_status_dtype = np.uint8
seg_initpoint_dtype = np.uint8
seg_endpoint_dtype = np.uint8
istate_type_dtype = np.uint8
istate_status_dtype = np.uint8

summary_table_dtype = np.dtype(
    [
        ('n_particles', seg_id_dtype),  # Number of live trajectories in this iteration
        ('norm', weight_dtype),  # Norm of probability, to watch for errors or drift
        ('min_bin_prob', weight_dtype),  # Per-bin minimum probability
        ('max_bin_prob', weight_dtype),  # Per-bin maximum probability
        ('min_seg_prob', weight_dtype),  # Per-segment minimum probability
        ('max_seg_prob', weight_dtype),  # Per-segment maximum probability
        ('cputime', utime_dtype),  # Total CPU time for this iteration
        ('walltime', utime_dtype),  # Total wallclock time for this iteration
        ('binhash', binhash_dtype),
    ]
)


# The HDF5 file tracks two distinct, but related, histories:
#    (1) the evolution of the trajectory, which requires only an identifier
#        of where a segment's initial state comes from (the "history graph");
#        this is stored as the parent_id field of the seg index
#    (2) the flow of probability due to splits, merges, and recycling events,
#        which can be thought of as an adjacency list (the "weight graph")
# segment ID is implied by the row in the index table, and so is not stored
# initpoint_type remains implicitly stored as negative IDs (if parent_id < 0, then init_state_id = -(parent_id+1)
seg_index_dtype = np.dtype(
    [
        ('weight', weight_dtype),  # Statistical weight of this segment
        ('parent_id', seg_id_dtype),  # ID of parent (for trajectory history)
        ('wtg_n_parents', np.uint),  # number of parents this segment has in the weight transfer graph
        ('wtg_offset', np.uint),  # offset into the weight transfer graph dataset
        ('cputime', utime_dtype),  # CPU time used in propagating this segment
        ('walltime', utime_dtype),  # Wallclock time used in propagating this segment
        ('endpoint_type', seg_endpoint_dtype),  # Endpoint type (will continue, merged, or recycled)
        ('status', seg_status_dtype),  # Status of propagation of this segment
    ]
)

# Index to basis/initial states
ibstate_index_dtype = np.dtype([('iter_valid', np.uint), ('n_bstates', np.uint), ('group_ref', h5ref_dtype)])

# Basis state index type
bstate_dtype = np.dtype(
    [
        ('label', vstr_dtype),  # An optional descriptive label
        ('probability', weight_dtype),  # Probability that this state will be selected
        ('auxref', vstr_dtype),  # An optional auxiliar data reference
    ]
)

# Even when initial state generation is off and basis states are passed through directly, an initial state entry
# is created, as that allows precise tracing of the history of a given state in the most complex case of
# a new initial state for every new trajectory.
istate_dtype = np.dtype(
    [
        ('iter_created', np.uint),  # Iteration during which this state was generated (0 for at w_init)
        ('iter_used', np.uint),  # When this state was used to start a new trajectory
        ('basis_state_id', seg_id_dtype),  # Which basis state this state was generated from
        ('istate_type', istate_type_dtype),  # What type this initial state is (generated or basis)
        ('istate_status', istate_status_dtype),  # Whether this initial state is ready to go
        ('basis_auxref', vstr_dtype),  # for start states to point back to their original location
    ]
)

tstate_index_dtype = np.dtype(
    [('iter_valid', np.uint), ('n_states', np.uint), ('group_ref', h5ref_dtype)]  # Iteration when this state list is valid
)  # Reference to a group containing further data; this will be the
# null reference if there is no target state for that timeframe.
tstate_dtype = np.dtype([('label', vstr_dtype)])  # An optional descriptive label for this state

# Support for west.we_driver.NewWeightEntry
nw_source_dtype = np.uint8
nw_index_dtype = np.dtype(
    [
        ('source_type', nw_source_dtype),
        ('weight', weight_dtype),
        ('prev_seg_id', seg_id_dtype),
        ('target_state_id', seg_id_dtype),
        ('initial_state_id', seg_id_dtype),
    ]
)

# Storage of bin identities
binning_index_dtype = np.dtype([('hash', binhash_dtype), ('pickle_len', np.uint32)])


class WESTDataManager:
    """Data manager for assisiting the reading and writing of WEST data from/to HDF5 files."""

    # defaults for various options
    default_iter_prec = 8
    default_we_h5filename = 'west.h5'
    default_we_h5file_driver = None
    default_flush_period = 60

    # Compress any auxiliary dataset whose total size (across all segments) is more than 1MB
    default_aux_compression_threshold = 1048576

    # Bin data horizontal (second dimension) chunk size
    binning_hchunksize = 4096

    # Number of rows to retrieve during a table scan
    table_scan_chunksize = 1024

    def flushing_lock(self):
        return flushing_lock(self.lock, self.we_h5file)

    def expiring_flushing_lock(self):
        next_flush = self.last_flush + self.flush_period
        return expiring_flushing_lock(self.lock, self.flush_backing, next_flush)

    def process_config(self):
        config = self.rc.config

        for entry, type_ in [('iter_prec', int)]:
            config.require_type_if_present(['west', 'data', entry], type_)

        self.we_h5filename = config.get_path(['west', 'data', 'west_data_file'], default=self.default_we_h5filename)
        self.we_h5file_driver = config.get_choice(
            ['west', 'data', 'west_data_file_driver'],
            [None, 'sec2', 'family'],
            default=self.default_we_h5file_driver,
            value_transform=(lambda x: x.lower() if x else None),
        )
        self.iter_prec = config.get(['west', 'data', 'iter_prec'], self.default_iter_prec)
        self.aux_compression_threshold = config.get(
            ['west', 'data', 'aux_compression_threshold'], self.default_aux_compression_threshold
        )
        self.flush_period = config.get(['west', 'data', 'flush_period'], self.default_flush_period)
        self.iter_ref_h5_template = config.get(['west', 'data', 'data_refs', 'iteration'], None)
        self.store_h5 = self.iter_ref_h5_template is not None

        # Process dataset options
        dsopts_list = config.get(['west', 'data', 'datasets']) or []
        for dsopts in dsopts_list:
            dsopts = normalize_dataset_options(dsopts, path_prefix='auxdata' if dsopts['name'] != 'pcoord' else '')
            try:
                self.dataset_options[dsopts['name']].update(dsopts)
            except KeyError:
                self.dataset_options[dsopts['name']] = dsopts

        if 'pcoord' in self.dataset_options:
            if self.dataset_options['pcoord']['h5path'] != 'pcoord':
                raise ValueError('cannot override pcoord storage location')

    def __init__(self, rc=None):
        self.rc = rc or westpa.rc

        self.we_h5filename = self.default_we_h5filename
        self.we_h5file_driver = self.default_we_h5file_driver
        self.we_h5file_version = None
        self.h5_access_mode = 'r+'
        self.iter_prec = self.default_iter_prec
        self.aux_compression_threshold = self.default_aux_compression_threshold

        self.we_h5file = None

        self.lock = threading.RLock()
        self.flush_period = None
        self.last_flush = 0

        self._system = None
        self.iter_ref_h5_template = None
        self.store_h5 = False

        self.dataset_options = {}
        self.process_config()

    @property
    def system(self):
        if self._system is None:
            self._system = self.rc.get_system_driver()
        return self._system

    @system.setter
    def system(self, system):
        self._system = system

    @property
    def closed(self):
        return self.we_h5file is None

    def iter_group_name(self, n_iter, absolute=True):
        if absolute:
            return '/iterations/iter_{:0{prec}d}'.format(int(n_iter), prec=self.iter_prec)
        else:
            return 'iter_{:0{prec}d}'.format(int(n_iter), prec=self.iter_prec)

    def require_iter_group(self, n_iter):
        '''Get the group associated with n_iter, creating it if necessary.'''
        with self.lock:
            iter_group = self.we_h5file.require_group('/iterations/iter_{:0{prec}d}'.format(int(n_iter), prec=self.iter_prec))
            iter_group.attrs['n_iter'] = n_iter
        return iter_group

    def del_iter_group(self, n_iter):
        with self.lock:
            del self.we_h5file['/iterations/iter_{:0{prec}d}'.format(int(n_iter), prec=self.iter_prec)]

    def get_iter_group(self, n_iter):
        with self.lock:
            try:
                return self.we_h5file['/iterations/iter_{:0{prec}d}'.format(int(n_iter), prec=self.iter_prec)]
            except KeyError:
                return self.we_h5file['/iter_{:0{prec}d}'.format(int(n_iter), prec=self.iter_prec)]

    def get_seg_index(self, n_iter):
        with self.lock:
            seg_index = self.get_iter_group(n_iter)['seg_index']
            return seg_index

    @property
    def current_iteration(self):
        with self.lock:
            h5file_attrs = self.we_h5file['/'].attrs
            h5file_attr_keys = list(h5file_attrs.keys())

            if 'west_current_iteration' in h5file_attr_keys:
                return int(self.we_h5file['/'].attrs['west_current_iteration'])
            else:
                return int(self.we_h5file['/'].attrs['wemd_current_iteration'])

    @current_iteration.setter
    def current_iteration(self, n_iter):
        with self.lock:
            self.we_h5file['/'].attrs['west_current_iteration'] = n_iter

    def open_backing(self, mode=None):
        '''Open the (already-created) HDF5 file named in self.west_h5filename.'''
        mode = mode or self.h5_access_mode
        if not self.we_h5file:
            log.debug('attempting to open {} with mode {}'.format(self.we_h5filename, mode))
            self.we_h5file = h5io.WESTPAH5File(self.we_h5filename, mode, driver=self.we_h5file_driver)

            h5file_attrs = self.we_h5file['/'].attrs
            h5file_attr_keys = list(h5file_attrs.keys())

            if 'west_iter_prec' in h5file_attr_keys:
                self.iter_prec = int(h5file_attrs['west_iter_prec'])
            elif 'wemd_iter_prec' in h5file_attr_keys:
                self.iter_prec = int(h5file_attrs['wemd_iter_prec'])
            else:
                log.info('iteration precision not stored in HDF5; using {:d}'.format(self.iter_prec))

            if 'west_file_format_version' in h5file_attr_keys:
                self.we_h5file_version = h5file_attrs['west_file_format_version']
            elif 'wemd_file_format_version' in h5file_attr_keys:
                self.we_h5file_version = h5file_attrs['wemd_file_format_version']
            else:
                log.info('WEST HDF5 file format version not stored, assuming 0')
                self.we_h5file_version = 0

            log.debug('opened WEST HDF5 file version {:d}'.format(self.we_h5file_version))

    def prepare_backing(self):  # istates):
        '''Create new HDF5 file'''
        self.we_h5file = h5py.File(self.we_h5filename, 'w', driver=self.we_h5file_driver)

        with self.flushing_lock():
            self.we_h5file['/'].attrs['west_file_format_version'] = file_format_version
            self.we_h5file['/'].attrs['west_iter_prec'] = self.iter_prec
            self.we_h5file['/'].attrs['west_version'] = westpa.__version__
            self.current_iteration = 0
            self.we_h5file['/'].create_dataset('summary', shape=(1,), dtype=summary_table_dtype, maxshape=(None,))
            self.we_h5file.create_group('/iterations')

    def close_backing(self):
        if self.we_h5file is not None:
            with self.lock:
                self.we_h5file.close()
            self.we_h5file = None

    def flush_backing(self):
        if self.we_h5file is not None:
            with self.lock:
                self.we_h5file.flush()
                self.last_flush = time.time()

    def save_target_states(self, tstates, n_iter=None):
        '''Save the given target states in the HDF5 file; they will be used for the next iteration to
        be propagated.  A complete set is required, even if nominally appending to an existing set,
        which simplifies the mapping of IDs to the table.'''

        system = self.system

        n_iter = n_iter or self.current_iteration

        # Assemble all the important data before we start to modify the HDF5 file
        tstates = list(tstates)
        if tstates:
            state_table = np.empty((len(tstates),), dtype=tstate_dtype)
            state_pcoords = np.empty((len(tstates), system.pcoord_ndim), dtype=system.pcoord_dtype)
            for i, state in enumerate(tstates):
                state.state_id = i
                state_table[i]['label'] = state.label
                state_pcoords[i] = state.pcoord
        else:
            state_table = None
            state_pcoords = None

        # Commit changes to HDF5
        with self.lock:
            master_group = self.we_h5file.require_group('tstates')

            try:
                master_index = master_group['index']
            except KeyError:
                master_index = master_group.create_dataset('index', shape=(1,), maxshape=(None,), dtype=tstate_index_dtype)
                n_sets = 1
            else:
                n_sets = len(master_index) + 1
                master_index.resize((n_sets,))

            set_id = n_sets - 1
            master_index_row = master_index[set_id]
            master_index_row['iter_valid'] = n_iter
            master_index_row['n_states'] = len(tstates)

            if tstates:
                state_group = master_group.create_group(str(set_id))
                master_index_row['group_ref'] = state_group.ref
                state_group['index'] = state_table
                state_group['pcoord'] = state_pcoords
            else:
                master_index_row['group_ref'] = None

            master_index[set_id] = master_index_row

    def _find_multi_iter_group(self, n_iter, master_group_name):
        with self.lock:
            master_group = self.we_h5file[master_group_name]
            master_index = master_group['index'][...]
            set_id = np.digitize([n_iter], master_index['iter_valid']) - 1
            group_ref = master_index[set_id]['group_ref']

            # Check if reference is Null
            if not bool(group_ref):
                return None

            # This extra [0] is to work around a bug in h5py
            try:
                group = self.we_h5file[group_ref]
            except (TypeError, AttributeError):
                group = self.we_h5file[group_ref[0]]
            else:
                log.debug('h5py fixed; remove alternate code path')
            log.debug('reference {!r} points to group {!r}'.format(group_ref, group))
            return group

    def find_tstate_group(self, n_iter):
        return self._find_multi_iter_group(n_iter, 'tstates')

    def find_ibstate_group(self, n_iter):
        return self._find_multi_iter_group(n_iter, 'ibstates')

    def get_target_states(self, n_iter):
        '''Return a list of Target objects representing the target (sink) states that are in use for iteration n_iter.
        Future iterations are assumed to continue from the most recent set of states.'''

        with self.lock:
            tstate_group = self.find_tstate_group(n_iter)

            if tstate_group is not None:
                tstate_index = tstate_group['index'][...]
                tstate_pcoords = tstate_group['pcoord'][...]

                tstates = [
                    TargetState(state_id=i, label=h5io.tostr(row['label']), pcoord=pcoord.copy())
                    for (i, (row, pcoord)) in enumerate(zip(tstate_index, tstate_pcoords))
                ]
            else:
                tstates = []

            return tstates

    def create_ibstate_group(self, basis_states, n_iter=None):
        '''Create the group used to store basis states and initial states (whose definitions are always
        coupled).  This group is hard-linked into all iteration groups that use these basis and
        initial states.'''

        with self.lock:
            n_iter = n_iter or self.current_iteration
            master_group = self.we_h5file.require_group('ibstates')

            try:
                master_index = master_group['index']
            except KeyError:
                master_index = master_group.create_dataset('index', dtype=ibstate_index_dtype, shape=(1,), maxshape=(None,))
                n_sets = 1
            else:
                n_sets = len(master_index) + 1
                master_index.resize((n_sets,))

            set_id = n_sets - 1
            master_index_row = master_index[set_id]
            master_index_row['iter_valid'] = n_iter
            master_index_row['n_bstates'] = len(basis_states)
            state_group = master_group.create_group(str(set_id))
            master_index_row['group_ref'] = state_group.ref

            if basis_states:
                system = self.system
                state_table = np.empty((len(basis_states),), dtype=bstate_dtype)
                state_pcoords = np.empty((len(basis_states), system.pcoord_ndim), dtype=system.pcoord_dtype)
                for i, state in enumerate(basis_states):
                    state.state_id = i
                    state_table[i]['label'] = state.label
                    state_table[i]['probability'] = state.probability
                    state_table[i]['auxref'] = state.auxref or ''
                    state_pcoords[i] = state.pcoord

                state_group['bstate_index'] = state_table
                state_group['bstate_pcoord'] = state_pcoords

            master_index[set_id] = master_index_row
            return state_group

    def create_ibstate_iter_h5file(self, basis_states):
        '''Create the per-iteration HDF5 file for the basis states (i.e., iteration 0).
        This special treatment is needed so that the analysis tools can access basis states
        more easily.'''

        if not self.store_h5:
            return

        segments = []
        for i, state in enumerate(basis_states):
            dummy_segment = Segment(
                n_iter=0,
                seg_id=state.state_id,
                parent_id=-(state.state_id + 1),
                weight=state.probability,
                wtg_parent_ids=None,
                pcoord=state.pcoord,
                status=Segment.SEG_STATUS_UNSET,
                data=state.data,
            )
            segments.append(dummy_segment)

        # # link the iteration file in west.h5
        self.prepare_iteration(0, segments)
        self.update_iter_h5file(0, segments)

    def update_iter_h5file(self, n_iter, segments):
        '''Write out the per-iteration HDF5 file with given segments and add an external link to it
        in the main HDF5 file (west.h5) if the link is not present.'''

        if not self.store_h5:
            return

        west_h5_file = makepath(self.we_h5filename)
        iter_ref_h5_file = makepath(self.iter_ref_h5_template, {'n_iter': n_iter})
        iter_ref_rel_path = relpath(iter_ref_h5_file, dirname(west_h5_file))

        with h5io.WESTIterationFile(iter_ref_h5_file, 'a') as outf:
            for segment in segments:
                outf.write_segment(segment, True)

        iter_group = self.get_iter_group(n_iter)

        if 'trajectories' not in iter_group:
            iter_group['trajectories'] = h5py.ExternalLink(iter_ref_rel_path, '/')

    def get_basis_states(self, n_iter=None):
        '''Return a list of BasisState objects representing the basis states that are in use for iteration n_iter.'''

        with self.lock:
            n_iter = n_iter or self.current_iteration
            ibstate_group = self.find_ibstate_group(n_iter)
            try:
                bstate_index = ibstate_group['bstate_index'][...]
            except KeyError:
                return []
            bstate_pcoords = ibstate_group['bstate_pcoord'][...]
            bstates = [
                BasisState(
                    state_id=i,
                    label=h5io.tostr(row['label']),
                    probability=row['probability'],
                    auxref=h5io.tostr(row['auxref']) or None,
                    pcoord=pcoord.copy(),
                )
                for (i, (row, pcoord)) in enumerate(zip(bstate_index, bstate_pcoords))
            ]

            bstate_total_prob = sum(bstate.probability for bstate in bstates)

            # This should run once in the second iteration, and only if start-states are specified,
            # but is necessary to re-normalize (i.e. normalize without start-state probabilities included)
            for i, bstate in enumerate(bstates):
                bstate.probability /= bstate_total_prob
                bstates[i] = bstate
            return bstates

    def create_initial_states(self, n_states, n_iter=None):
        '''Create storage for ``n_states`` initial states associated with iteration ``n_iter``, and
        return bare InitialState objects with only state_id set.'''

        system = self.system
        with self.lock:
            n_iter = n_iter or self.current_iteration
            ibstate_group = self.find_ibstate_group(n_iter)

            try:
                istate_index = ibstate_group['istate_index']
            except KeyError:
                istate_index = ibstate_group.create_dataset('istate_index', dtype=istate_dtype, shape=(n_states,), maxshape=(None,))
                istate_pcoords = ibstate_group.create_dataset(
                    'istate_pcoord',
                    dtype=system.pcoord_dtype,
                    shape=(n_states, system.pcoord_ndim),
                    maxshape=(None, system.pcoord_ndim),
                )
                len_index = len(istate_index)
                first_id = 0
            else:
                first_id = len(istate_index)
                len_index = len(istate_index) + n_states
                istate_index.resize((len_index,))
                istate_pcoords = ibstate_group['istate_pcoord']
                istate_pcoords.resize((len_index, system.pcoord_ndim))

        index_entries = istate_index[first_id:len_index]
        new_istates = []
        for irow, row in enumerate(index_entries):
            row['iter_created'] = n_iter
            row['istate_status'] = InitialState.ISTATE_STATUS_PENDING
            new_istates.append(
                InitialState(
                    state_id=first_id + irow,
                    basis_state_id=None,
                    iter_created=n_iter,
                    istate_status=InitialState.ISTATE_STATUS_PENDING,
                )
            )
        istate_index[first_id:len_index] = index_entries
        return new_istates

    def update_initial_states(self, initial_states, n_iter=None):
        '''Save the given initial states in the HDF5 file'''

        system = self.system
        initial_states = sorted(initial_states, key=attrgetter('state_id'))
        if not initial_states:
            return

        with self.lock:
            n_iter = n_iter or self.current_iteration
            ibstate_group = self.find_ibstate_group(n_iter)
            state_ids = [state.state_id for state in initial_states]
            index_entries = ibstate_group['istate_index'][state_ids]
            pcoord_vals = np.empty((len(initial_states), system.pcoord_ndim), dtype=system.pcoord_dtype)
            for i, initial_state in enumerate(initial_states):
                index_entries[i]['iter_created'] = initial_state.iter_created
                index_entries[i]['iter_used'] = initial_state.iter_used or InitialState.ISTATE_UNUSED
                index_entries[i]['basis_state_id'] = (
                    initial_state.basis_state_id if initial_state.basis_state_id is not None else -1
                )
                index_entries[i]['istate_type'] = initial_state.istate_type or InitialState.ISTATE_TYPE_UNSET
                index_entries[i]['istate_status'] = initial_state.istate_status or InitialState.ISTATE_STATUS_PENDING
                pcoord_vals[i] = initial_state.pcoord

                index_entries[i]['basis_auxref'] = initial_state.basis_auxref or ""

            ibstate_group['istate_index'][state_ids] = index_entries
            ibstate_group['istate_pcoord'][state_ids] = pcoord_vals

            if self.store_h5:
                segments = []
                for i, state in enumerate(initial_states):
                    dummy_segment = Segment(
                        n_iter=-state.iter_created,
                        seg_id=state.state_id,
                        parent_id=state.basis_state_id,
                        wtg_parent_ids=None,
                        pcoord=state.pcoord,
                        status=Segment.SEG_STATUS_PREPARED,
                        data=state.data,
                    )
                    segments.append(dummy_segment)
                self.update_iter_h5file(0, segments)

    def get_initial_states(self, n_iter=None):
        states = []
        with self.lock:
            n_iter = n_iter or self.current_iteration
            ibstate_group = self.find_ibstate_group(n_iter)
            try:
                istate_index = ibstate_group['istate_index'][...]
            except KeyError:
                return []
            istate_pcoords = ibstate_group['istate_pcoord'][...]

            for state_id, (state, pcoord) in enumerate(zip(istate_index, istate_pcoords)):
                states.append(
                    InitialState(
                        state_id=state_id,
                        basis_state_id=int(state['basis_state_id']),
                        iter_created=int(state['iter_created']),
                        iter_used=int(state['iter_used']),
                        istate_type=int(state['istate_type']),
                        basis_auxref=h5io.tostr(state['basis_auxref']),
                        pcoord=pcoord.copy(),
                    )
                )
            return states

    def get_segment_initial_states(self, segments, n_iter=None):
        '''Retrieve all initial states referenced by the given segments.'''

        with self.lock:
            n_iter = n_iter or self.current_iteration
            ibstate_group = self.get_iter_group(n_iter)['ibstates']

            istate_ids = {-int(segment.parent_id + 1) for segment in segments if segment.parent_id < 0}
            sorted_istate_ids = sorted(istate_ids)
            if not sorted_istate_ids:
                return []

            istate_rows = ibstate_group['istate_index'][sorted_istate_ids][...]
            istate_pcoords = ibstate_group['istate_pcoord'][sorted_istate_ids][...]
            istates = []

            for state_id, state, pcoord in zip(sorted_istate_ids, istate_rows, istate_pcoords):
                try:
                    b_auxref = h5io.tostr(state['basis_auxref'])
                except ValueError:
                    b_auxref = ''
                istate = InitialState(
                    state_id=state_id,
                    basis_state_id=int(state['basis_state_id']),
                    iter_created=int(state['iter_created']),
                    iter_used=int(state['iter_used']),
                    istate_type=int(state['istate_type']),
                    basis_auxref=b_auxref,
                    pcoord=pcoord.copy(),
                )
                istates.append(istate)
            return istates

    def get_unused_initial_states(self, n_states=None, n_iter=None):
        '''Retrieve any prepared but unused initial states applicable to the given iteration.
        Up to ``n_states`` states are returned; if ``n_states`` is None, then all unused states
        are returned.'''

        n_states = n_states or sys.maxsize
        ISTATE_UNUSED = InitialState.ISTATE_UNUSED
        ISTATE_STATUS_PREPARED = InitialState.ISTATE_STATUS_PREPARED
        with self.lock:
            n_iter = n_iter or self.current_iteration
            ibstate_group = self.find_ibstate_group(n_iter)
            istate_index = ibstate_group['istate_index']
            istate_pcoords = ibstate_group['istate_pcoord']
            n_index_entries = istate_index.len()
            chunksize = self.table_scan_chunksize

            states = []
            istart = 0
            while istart < n_index_entries and len(states) < n_states:
                istop = min(istart + chunksize, n_index_entries)
                istate_chunk = istate_index[istart:istop]
                pcoord_chunk = istate_pcoords[istart:istop]
                # state_ids = np.arange(istart,istop,dtype=np.uint)

                for ci in range(len(istate_chunk)):
                    row = istate_chunk[ci]
                    pcoord = pcoord_chunk[ci]
                    state_id = istart + ci
                    if row['iter_used'] == ISTATE_UNUSED and row['istate_status'] == ISTATE_STATUS_PREPARED:
                        istate = InitialState(
                            state_id=state_id,
                            basis_state_id=int(row['basis_state_id']),
                            iter_created=int(row['iter_created']),
                            iter_used=0,
                            istate_type=int(row['istate_type']),
                            pcoord=pcoord.copy(),
                            istate_status=ISTATE_STATUS_PREPARED,
                        )
                        states.append(istate)
                    del row, pcoord, state_id
                istart += chunksize
                del istate_chunk, pcoord_chunk  # , state_ids, unused, ids_of_unused
            log.debug('found {:d} unused states'.format(len(states)))
            return states[:n_states]

    def prepare_iteration(self, n_iter, segments):
        """Prepare for a new iteration by creating space to store the new iteration's data.
        The number of segments, their IDs, and their lineage must be determined and included
        in the set of segments passed in."""

        log.debug('preparing HDF5 group for iteration %d (%d segments)' % (n_iter, len(segments)))

        # Ensure we have a list for guaranteed ordering
        init = n_iter == 0
        segments = list(segments)
        n_particles = len(segments)
        system = self.system
        pcoord_ndim = system.pcoord_ndim
        pcoord_len = 2 if init else system.pcoord_len
        pcoord_dtype = system.pcoord_dtype

        with self.lock:
            if not init:
                # Create a table of summary information about each iteration
                summary_table = self.we_h5file['summary']
                if len(summary_table) < n_iter:
                    summary_table.resize((n_iter + 1,))

            iter_group = self.require_iter_group(n_iter)

            for linkname in ('seg_index', 'pcoord', 'wtgraph'):
                try:
                    del iter_group[linkname]
                except KeyError:
                    pass

            # everything indexed by [particle] goes in an index table
            seg_index_table_ds = iter_group.create_dataset('seg_index', shape=(n_particles,), dtype=seg_index_dtype)
            # unfortunately, h5py doesn't like in-place modification of individual fields; it expects
            # tuples. So, construct everything in a numpy array and then dump the whole thing into hdf5
            # In fact, this appears to be an h5py best practice (collect as much in ram as possible and then dump)
            seg_index_table = seg_index_table_ds[...]

            if not init:
                summary_row = np.zeros((1,), dtype=summary_table_dtype)
                summary_row['n_particles'] = n_particles
                summary_row['norm'] = np.add.reduce(list(map(attrgetter('weight'), segments)))
                summary_table[n_iter - 1] = summary_row

            # pcoord is indexed as [particle, time, dimension]
            pcoord_opts = self.dataset_options.get('pcoord', {'name': 'pcoord', 'h5path': 'pcoord', 'compression': False})
            shape = (n_particles, pcoord_len, pcoord_ndim)
            pcoord_ds = create_dataset_from_dsopts(iter_group, pcoord_opts, shape, pcoord_dtype)
            pcoord = np.empty((n_particles, pcoord_len, pcoord_ndim), pcoord_dtype)

            total_parents = 0
            for seg_id, segment in enumerate(segments):
                if segment.seg_id is not None:
                    assert segment.seg_id == seg_id
                else:
                    segment.seg_id = seg_id
                # Parent must be set, though what it means depends on initpoint_type
                assert segment.parent_id is not None
                segment.seg_id = seg_id
                seg_index_table[seg_id]['status'] = segment.status
                seg_index_table[seg_id]['weight'] = segment.weight
                seg_index_table[seg_id]['parent_id'] = segment.parent_id
                seg_index_table[seg_id]['wtg_n_parents'] = len(segment.wtg_parent_ids)
                seg_index_table[seg_id]['wtg_offset'] = total_parents
                total_parents += len(segment.wtg_parent_ids)

                # Assign progress coordinate if any exists
                if segment.pcoord is not None:
                    if init:
                        if segment.pcoord.shape != pcoord.shape[2:]:
                            raise ValueError(
                                'basis state pcoord shape [%r] does not match expected shape [%r]'
                                % (segment.pcoord.shape, pcoord.shape[2:])
                            )
                        # Initial pcoord
                        pcoord[seg_id, 1, :] = segment.pcoord[:]
                    else:
                        if len(segment.pcoord) == 1:
                            # Initial pcoord
                            pcoord[seg_id, 0, :] = segment.pcoord[0, :]
                        elif segment.pcoord.shape != pcoord.shape[1:]:
                            raise ValueError(
                                'segment pcoord shape [%r] does not match expected shape [%r]'
                                % (segment.pcoord.shape, pcoord.shape[1:])
                            )
                        else:
                            pcoord[seg_id, ...] = segment.pcoord

            if total_parents > 0:
                wtgraph_ds = iter_group.create_dataset('wtgraph', (total_parents,), seg_id_dtype, compression='gzip', shuffle=True)
                parents = np.empty((total_parents,), seg_id_dtype)

                for seg_id, segment in enumerate(segments):
                    offset = seg_index_table[seg_id]['wtg_offset']
                    extent = seg_index_table[seg_id]['wtg_n_parents']
                    parent_list = list(segment.wtg_parent_ids)
                    parents[offset : offset + extent] = parent_list[:]

                    assert set(parents[offset : offset + extent]) == set(segment.wtg_parent_ids)

                wtgraph_ds[:] = parents

            # Create convenient hard links
            self.update_iter_group_links(n_iter)

            # Since we accumulated many of these changes in RAM (and not directly in HDF5), propagate
            # the changes out to HDF5
            seg_index_table_ds[:] = seg_index_table
            pcoord_ds[...] = pcoord

    def update_iter_group_links(self, n_iter):
        '''Update the per-iteration hard links pointing to the tables of target and initial/basis states for the
        given iteration.  These links are not used by this class, but are remarkably convenient for third-party
        analysis tools and hdfview.'''

        with self.lock:
            iter_group = self.require_iter_group(n_iter)

            for linkname in ('ibstates', 'tstates'):
                try:
                    del iter_group[linkname]
                except KeyError:
                    pass

            iter_group['ibstates'] = self.find_ibstate_group(n_iter)

            tstate_group = self.find_tstate_group(n_iter)
            if tstate_group is not None:
                iter_group['tstates'] = tstate_group

    def get_iter_summary(self, n_iter=None):
        n_iter = n_iter or self.current_iteration
        with self.lock:
            return self.we_h5file['summary'][n_iter - 1]

    def update_iter_summary(self, summary, n_iter=None):
        n_iter = n_iter or self.current_iteration
        with self.lock:
            self.we_h5file['summary'][n_iter - 1] = summary

    def del_iter_summary(self, min_iter):  # delete the iterations starting at min_iter
        with self.lock:
            self.we_h5file['summary'].resize((min_iter - 1,))

    def update_segments(self, n_iter, segments):
        '''Update segment information in the HDF5 file; all prior information for each
        ``segment`` is overwritten, except for parent and weight transfer information.'''

        segments = sorted(segments, key=attrgetter('seg_id'))

        with self.lock:
            iter_group = self.get_iter_group(n_iter)

            pc_dsid = iter_group['pcoord'].id
            si_dsid = iter_group['seg_index'].id

            seg_ids = [segment.seg_id for segment in segments]
            n_segments = len(segments)
            n_total_segments = si_dsid.shape[0]
            system = self.system
            pcoord_ndim = system.pcoord_ndim
            pcoord_len = system.pcoord_len
            pcoord_dtype = system.pcoord_dtype

            seg_index_entries = np.empty((n_segments,), dtype=seg_index_dtype)
            pcoord_entries = np.empty((n_segments, pcoord_len, pcoord_ndim), dtype=pcoord_dtype)

            pc_msel = h5s.create_simple(pcoord_entries.shape, (h5s.UNLIMITED,) * pcoord_entries.ndim)
            pc_msel.select_all()
            si_msel = h5s.create_simple(seg_index_entries.shape, (h5s.UNLIMITED,))
            si_msel.select_all()
            pc_fsel = pc_dsid.get_space()
            si_fsel = si_dsid.get_space()

            for iseg in range(n_segments):
                seg_id = seg_ids[iseg]
                op = h5s.SELECT_OR if iseg != 0 else h5s.SELECT_SET
                si_fsel.select_hyperslab((seg_id,), (1,), op=op)
                pc_fsel.select_hyperslab((seg_id, 0, 0), (1, pcoord_len, pcoord_ndim), op=op)

            # read summary data so that we have valud parent and weight transfer information
            si_dsid.read(si_msel, si_fsel, seg_index_entries)

            for iseg, (segment, ientry) in enumerate(zip(segments, seg_index_entries)):
                ientry['status'] = segment.status
                ientry['endpoint_type'] = segment.endpoint_type or Segment.SEG_ENDPOINT_UNSET
                ientry['cputime'] = segment.cputime
                ientry['walltime'] = segment.walltime
                ientry['weight'] = segment.weight

                pcoord_entries[iseg] = segment.pcoord

            # write progress coordinates and index using low level HDF5 functions for efficiency
            si_dsid.write(si_msel, si_fsel, seg_index_entries)
            pc_dsid.write(pc_msel, pc_fsel, pcoord_entries)

            # Now, to deal with auxiliary data
            # If any segment has any auxiliary data, then the aux dataset must spring into
            # existence. Each is named according to the name in segment.data, and has shape
            # (n_total_segs, ...) where the ... is the shape of the data in segment.data (and may be empty
            # in the case of scalar data) and dtype is taken from the data type of the data entry
            # compression is on by default for datasets that will be more than 1MiB

            # a mapping of data set name to (per-segment shape, data type) tuples
            dsets = {}

            # First we scan for presence, shape, and data type of auxiliary data sets
            for segment in segments:
                if segment.data:
                    for dsname in segment.data:
                        if dsname.startswith('iterh5/'):
                            continue
                        data = np.asarray(segment.data[dsname], order='C')
                        segment.data[dsname] = data
                        dsets[dsname] = (data.shape, data.dtype)

            # Then we iterate over data sets and store data
            if dsets:
                for dsname, (shape, dtype) in dsets.items():
                    # dset = self._require_aux_dataset(iter_group, dsname, n_total_segments, shape, dtype)
                    try:
                        dsopts = self.dataset_options[dsname]
                    except KeyError:
                        dsopts = normalize_dataset_options({'name': dsname}, path_prefix='auxdata')

                    shape = (n_total_segments,) + shape
                    dset = require_dataset_from_dsopts(
                        iter_group, dsopts, shape, dtype, autocompress_threshold=self.aux_compression_threshold, n_iter=n_iter
                    )
                    if dset is None:
                        # storage is suppressed
                        continue
                    for segment in segments:
                        try:
                            auxdataset = segment.data[dsname]
                        except KeyError:
                            pass
                        else:
                            source_rank = len(auxdataset.shape)
                            source_sel = h5s.create_simple(auxdataset.shape, (h5s.UNLIMITED,) * source_rank)
                            source_sel.select_all()
                            dest_sel = dset.id.get_space()
                            dest_sel.select_hyperslab((segment.seg_id,) + (0,) * source_rank, (1,) + auxdataset.shape)
                            dset.id.write(source_sel, dest_sel, auxdataset)
                    if 'delram' in list(dsopts.keys()):
                        del dsets[dsname]

            self.update_iter_h5file(n_iter, segments)

    def get_segments(self, n_iter=None, seg_ids=None, load_pcoords=True):
        '''Return the given (or all) segments from a given iteration.

        If the optional parameter ``load_auxdata`` is true, then all auxiliary datasets
        available are loaded and mapped onto the ``data`` dictionary of each segment. If
        ``load_auxdata`` is None, then use the default ``self.auto_load_auxdata``, which can
        be set by the option ``load_auxdata`` in the ``[data]`` section of ``west.cfg``. This
        essentially requires as much RAM as there is per-iteration auxiliary data, so this
        behavior is not on by default.'''

        n_iter = n_iter or self.current_iteration
        file_version = self.we_h5file_version

        with self.lock:
            iter_group = self.get_iter_group(n_iter)
            seg_index_ds = iter_group['seg_index']

            if file_version < 5:
                all_parent_ids = iter_group['parents'][...]
            else:
                all_parent_ids = iter_group['wtgraph'][...]

            if seg_ids is not None:
                seg_ids = list(sorted(seg_ids))
                seg_index_entries = seg_index_ds[seg_ids]
                if load_pcoords:
                    pcoord_entries = iter_group['pcoord'][seg_ids]
            else:
                seg_ids = list(range(len(seg_index_ds)))
                seg_index_entries = seg_index_ds[...]
                if load_pcoords:
                    pcoord_entries = iter_group['pcoord'][...]

            segments = []

            for iseg, (seg_id, row) in enumerate(zip(seg_ids, seg_index_entries)):
                segment = Segment(
                    seg_id=seg_id,
                    n_iter=n_iter,
                    status=int(row['status']),
                    endpoint_type=int(row['endpoint_type']),
                    walltime=float(row['walltime']),
                    cputime=float(row['cputime']),
                    weight=float(row['weight']),
                )

                if load_pcoords:
                    segment.pcoord = pcoord_entries[iseg]

                if file_version < 5:
                    wtg_n_parents = row['n_parents']
                    wtg_offset = row['parents_offset']
                    wtg_parent_ids = all_parent_ids[wtg_offset : wtg_offset + wtg_n_parents]
                    segment.parent_id = int(wtg_parent_ids[0])
                else:
                    wtg_n_parents = row['wtg_n_parents']
                    wtg_offset = row['wtg_offset']
                    wtg_parent_ids = all_parent_ids[wtg_offset : wtg_offset + wtg_n_parents]
                    segment.parent_id = int(row['parent_id'])
                segment.wtg_parent_ids = set(map(int, wtg_parent_ids))
                assert len(segment.wtg_parent_ids) == wtg_n_parents
                segments.append(segment)
            del all_parent_ids
            if load_pcoords:
                del pcoord_entries

            # If any other data sets are requested, load them as well
            for dsinfo in self.dataset_options.values():
                if dsinfo.get('load', False):
                    dsname = dsinfo['name']
                    try:
                        ds = iter_group[dsinfo['h5path']]
                    except KeyError:
                        ds = None

                    if ds is not None:
                        for segment in segments:
                            seg_id = segment.seg_id
                            segment.data[dsname] = ds[seg_id]

        return segments

    def prepare_segment_restarts(self, segments, basis_states=None, initial_states=None):
        '''Prepare the necessary folder and files given the data stored in parent per-iteration HDF5 file
        for propagating the simulation. ``basis_states`` and ``initial_states`` should be provided if the
        segments are newly created'''

        if not self.store_h5:
            return

        for segment in segments:
            if segment.parent_id < 0:
                if initial_states is None or basis_states is None:
                    raise ValueError('initial and basis states required for preparing the segments')
                initial_state = initial_states[segment.initial_state_id]
                # Check if it's a start state
                if initial_state.istate_type == InitialState.ISTATE_TYPE_START:
                    log.debug(
                        f'Skip reading start state file from per-iteration HDF5 file for initial state {segment.initial_state_id}'
                    )
                    continue
                else:
                    basis_state = basis_states[initial_state.basis_state_id]

                parent = Segment(n_iter=0, seg_id=basis_state.state_id)
            else:
                parent = Segment(n_iter=segment.n_iter - 1, seg_id=segment.parent_id)

            try:
                parent_iter_ref_h5_file = makepath(self.iter_ref_h5_template, {'n_iter': parent.n_iter})

                with h5io.WESTIterationFile(parent_iter_ref_h5_file, 'r') as outf:
                    outf.read_restart(parent)

                segment.data['iterh5/restart'] = parent.data['iterh5/restart']
            except Exception as e:
                print('could not prepare restart data for segment {}/{}: {}'.format(segment.n_iter, segment.seg_id, str(e)))

    def get_all_parent_ids(self, n_iter):
        file_version = self.we_h5file_version
        with self.lock:
            iter_group = self.get_iter_group(n_iter)
            seg_index = iter_group['seg_index']

            if file_version < 5:
                offsets = seg_index['parents_offset']
                all_parents = iter_group['parents'][...]
                return all_parents.take(offsets)
            else:
                return seg_index['parent_id']

    def get_parent_ids(self, n_iter, seg_ids=None):
        '''Return a sequence of the parent IDs of the given seg_ids.'''

        file_version = self.we_h5file_version

        with self.lock:
            iter_group = self.get_iter_group(n_iter)
            seg_index = iter_group['seg_index']

            if seg_ids is None:
                seg_ids = range(len(seg_index))

            if file_version < 5:
                offsets = seg_index['parents_offset']
                all_parents = iter_group['parents'][...]
                return [all_parents[offsets[seg_id]] for seg_id in seg_ids]
            else:
                all_parents = seg_index['parent_id']
                return [all_parents[seg_id] for seg_id in seg_ids]

    def get_weights(self, n_iter, seg_ids):
        '''Return the weights associated with the given seg_ids'''

        unique_ids = sorted(set(seg_ids))
        if not unique_ids:
            return []
        with self.lock:
            iter_group = self.get_iter_group(n_iter)
            index_subset = iter_group['seg_index'][unique_ids]
            weight_map = dict(zip(unique_ids, index_subset['weight']))
            return [weight_map[seg_id] for seg_id in seg_ids]

    def get_child_ids(self, n_iter, seg_id):
        '''Return the seg_ids of segments who have the given segment as a parent.'''

        with self.lock:
            if n_iter == self.current_iteration:
                return []

            iter_group = self.get_iter_group(n_iter + 1)
            seg_index = iter_group['seg_index']
            seg_ids = np.arange(len(seg_index), dtype=seg_id_dtype)

            if self.we_h5file_version < 5:
                offsets = seg_index['parents_offset']
                all_parent_ids = iter_group['parents'][...]
                parent_ids = np.array([all_parent_ids[offset] for offset in offsets])
            else:
                parent_ids = seg_index['parent_id']

            return seg_ids[parent_ids == seg_id]

    def get_children(self, segment):
        '''Return all segments which have the given segment as a parent'''

        if segment.n_iter == self.current_iteration:
            return []

        # Examine the segment index from the following iteration to see who has this segment
        # as a parent.  We don't need to worry about the number of parents each segment
        # has, since each has at least one, and indexing on the offset into the parents array
        # gives the primary parent ID

        with self.lock:
            iter_group = self.get_iter_group(segment.n_iter + 1)
            seg_index = iter_group['seg_index'][...]

            # This is one of the slowest pieces of code I've ever written...
            # seg_index = iter_group['seg_index'][...]
            # seg_ids = [seg_id for (seg_id,row) in enumerate(seg_index)
            #           if all_parent_ids[row['parents_offset']] == segment.seg_id]
            # return self.get_segments_by_id(segment.n_iter+1, seg_ids)
            if self.we_h5file_version < 5:
                parents = iter_group['parents'][seg_index['parent_offsets']]
            else:
                parents = seg_index['parent_id']
            all_seg_ids = np.arange(seg_index.len(), dtype=np.uintp)
            seg_ids = all_seg_ids[parents == segment.seg_id]
            # the above will return a scalar if only one is found, so convert
            # to a list if necessary
            try:
                len(seg_ids)
            except TypeError:
                seg_ids = [seg_ids]

            return self.get_segments(segment.n_iter + 1, seg_ids)

    # The following are dictated by the SimManager interface
    def prepare_run(self):
        self.open_backing()

    def finalize_run(self):
        self.flush_backing()
        self.close_backing()

    def save_new_weight_data(self, n_iter, new_weights):
        '''Save a set of NewWeightEntry objects to HDF5. Note that this should
        be called for the iteration in which the weights appear in their
        new locations (e.g. for recycled walkers, the iteration following
        recycling).'''

        if not new_weights:
            return

        system = self.system

        index = np.empty(len(new_weights), dtype=nw_index_dtype)
        prev_init_pcoords = system.new_pcoord_array(len(new_weights))
        prev_final_pcoords = system.new_pcoord_array(len(new_weights))
        new_init_pcoords = system.new_pcoord_array(len(new_weights))

        for ientry, nwentry in enumerate(new_weights):
            row = index[ientry]
            row['source_type'] = nwentry.source_type
            row['weight'] = nwentry.weight
            row['prev_seg_id'] = nwentry.prev_seg_id
            # the following use -1 as a sentinel for a missing value
            row['target_state_id'] = nwentry.target_state_id if nwentry.target_state_id is not None else -1
            row['initial_state_id'] = nwentry.initial_state_id if nwentry.initial_state_id is not None else -1

            index[ientry] = row

            if nwentry.prev_init_pcoord is not None:
                prev_init_pcoords[ientry] = nwentry.prev_init_pcoord

            if nwentry.prev_final_pcoord is not None:
                prev_final_pcoords[ientry] = nwentry.prev_final_pcoord

            if nwentry.new_init_pcoord is not None:
                new_init_pcoords[ientry] = nwentry.new_init_pcoord

        with self.lock:
            iter_group = self.get_iter_group(n_iter)
            try:
                del iter_group['new_weights']
            except KeyError:
                pass

            nwgroup = iter_group.create_group('new_weights')
            nwgroup['index'] = index
            nwgroup['prev_init_pcoord'] = prev_init_pcoords
            nwgroup['prev_final_pcoord'] = prev_final_pcoords
            nwgroup['new_init_pcoord'] = new_init_pcoords

    def get_new_weight_data(self, n_iter):
        with self.lock:
            iter_group = self.get_iter_group(n_iter)

            try:
                nwgroup = iter_group['new_weights']
            except KeyError:
                return []

            try:
                index = nwgroup['index'][...]
                prev_init_pcoords = nwgroup['prev_init_pcoord'][...]
                prev_final_pcoords = nwgroup['prev_final_pcoord'][...]
                new_init_pcoords = nwgroup['new_init_pcoord'][...]
            except (KeyError, ValueError):  # zero-length selections raise ValueError
                return []

        entries = []
        for i in range(len(index)):
            irow = index[i]

            prev_seg_id = irow['prev_seg_id']
            if prev_seg_id == -1:
                prev_seg_id = None

            initial_state_id = irow['initial_state_id']
            if initial_state_id == -1:
                initial_state_id = None

            target_state_id = irow['target_state_id']
            if target_state_id == -1:
                target_state_id = None

            entry = NewWeightEntry(
                source_type=irow['source_type'],
                weight=irow['weight'],
                prev_seg_id=prev_seg_id,
                prev_init_pcoord=prev_init_pcoords[i].copy(),
                prev_final_pcoord=prev_final_pcoords[i].copy(),
                new_init_pcoord=new_init_pcoords[i].copy(),
                target_state_id=target_state_id,
                initial_state_id=initial_state_id,
            )

            entries.append(entry)
        return entries

    def find_bin_mapper(self, hashval):
        '''Check to see if the given has value is in the binning table. Returns the index in the
        bin data tables if found, or raises KeyError if not.'''

        try:
            hashval = hashval.hexdigest()
        except AttributeError:
            pass

        with self.lock:
            # these will raise KeyError if the group doesn't exist, which also means
            # that bin data is not available, so no special treatment here
            try:
                binning_group = self.we_h5file['/bin_topologies']
                index = binning_group['index']
            except KeyError:
                raise KeyError('hash {} not found'.format(hashval))

            n_entries = len(index)
            if n_entries == 0:
                raise KeyError('hash {} not found'.format(hashval))

            chunksize = self.table_scan_chunksize
            for istart in range(0, n_entries, chunksize):
                chunk = index[istart : min(istart + chunksize, n_entries)]
                for i in range(len(chunk)):
                    if chunk[i]['hash'] == bytes(hashval, 'utf-8'):
                        return istart + i

            raise KeyError('hash {} not found'.format(hashval))

    def get_bin_mapper(self, hashval):
        '''Look up the given hash value in the binning table, unpickling and returning the corresponding
        bin mapper if available, or raising KeyError if not.'''

        # Convert to a hex digest if we need to
        try:
            hashval = hashval.hexdigest()
        except AttributeError:
            pass

        with self.lock:
            # these will raise KeyError if the group doesn't exist, which also means
            # that bin data is not available, so no special treatment here
            try:
                binning_group = self.we_h5file['/bin_topologies']
                index = binning_group['index']
                pkl = binning_group['pickles']
            except KeyError:
                raise KeyError('hash {} not found. Could not retrieve binning group'.format(hashval))

            n_entries = len(index)
            if n_entries == 0:
                raise KeyError('hash {} not found. No entries in index'.format(hashval))

            chunksize = self.table_scan_chunksize

            for istart in range(0, n_entries, chunksize):
                chunk = index[istart : min(istart + chunksize, n_entries)]
                for i in range(len(chunk)):
                    if chunk[i]['hash'] == bytes(hashval, 'utf-8'):
                        pkldat = bytes(pkl[istart + i, 0 : chunk[i]['pickle_len']].data)
                        mapper = pickle.loads(pkldat)
                        log.debug('loaded {!r} from {!r}'.format(mapper, binning_group))
                        log.debug('hash value {!r}'.format(hashval))
                        return mapper

            raise KeyError('hash {} not found'.format(hashval))

    def save_bin_mapper(self, hashval, pickle_data):
        '''Store the given mapper in the table of saved mappers. If the mapper cannot be stored,
        PickleError will be raised. Returns the index in the bin data tables where the mapper is stored.'''

        try:
            hashval = hashval.hexdigest()
        except AttributeError:
            pass
        pickle_data = bytes(pickle_data)

        # First, scan to see if the mapper already is in the HDF5 file
        try:
            return self.find_bin_mapper(hashval)
        except KeyError:
            pass

        # At this point, we have a valid pickle and know it's not stored
        with self.lock:
            binning_group = self.we_h5file.require_group('/bin_topologies')

            try:
                index = binning_group['index']
                pickle_ds = binning_group['pickles']
            except KeyError:
                index = binning_group.create_dataset('index', shape=(1,), maxshape=(None,), dtype=binning_index_dtype)
                pickle_ds = binning_group.create_dataset(
                    'pickles',
                    dtype=np.uint8,
                    shape=(1, len(pickle_data)),
                    maxshape=(None, None),
                    chunks=(1, 4096),
                    compression='gzip',
                    compression_opts=9,
                )
                n_entries = 1
            else:
                n_entries = len(index) + 1
                index.resize((n_entries,))
                new_hsize = max(pickle_ds.shape[1], len(pickle_data))
                pickle_ds.resize((n_entries, new_hsize))

            index_row = index[n_entries - 1]
            index_row['hash'] = hashval
            index_row['pickle_len'] = len(pickle_data)
            index[n_entries - 1] = index_row
            pickle_ds[n_entries - 1, : len(pickle_data)] = memoryview(pickle_data)
            return n_entries - 1

    def save_iter_binning(self, n_iter, hashval, pickled_mapper, target_counts):
        '''Save information about the binning used to generate segments for iteration n_iter.'''

        with self.lock:
            iter_group = self.get_iter_group(n_iter)

            try:
                del iter_group['bin_target_counts']
            except KeyError:
                pass

            iter_group['bin_target_counts'] = target_counts

            if hashval and pickled_mapper:
                self.save_bin_mapper(hashval, pickled_mapper)
                iter_group.attrs['binhash'] = hashval
            else:
                iter_group.attrs['binhash'] = ''


def normalize_dataset_options(dsopts, path_prefix='', n_iter=0):
    dsopts = dict(dsopts)

    ds_name = dsopts['name']
    if path_prefix:
        default_h5path = '{}/{}'.format(path_prefix, ds_name)
    else:
        default_h5path = ds_name

    dsopts.setdefault('h5path', default_h5path)
    dtype = dsopts.get('dtype')
    if dtype:
        if isinstance(dtype, str):
            try:
                dsopts['dtype'] = np.dtype(getattr(np, dtype))
            except AttributeError:
                dsopts['dtype'] = np.dtype(getattr(builtins, dtype))
        else:
            dsopts['dtype'] = np.dtype(dtype)

    dsopts['store'] = bool(dsopts['store']) if 'store' in dsopts else True
    dsopts['load'] = bool(dsopts['load']) if 'load' in dsopts else False

    return dsopts


def create_dataset_from_dsopts(group, dsopts, shape=None, dtype=None, data=None, autocompress_threshold=None, n_iter=None):
    # log.debug('create_dataset_from_dsopts(group={!r}, dsopts={!r}, shape={!r}, dtype={!r}, data={!r}, autocompress_threshold={!r})'
    #          .format(group,dsopts,shape,dtype,data,autocompress_threshold))
    if not dsopts.get('store', True):
        return None

    if 'file' in list(dsopts.keys()):
        import h5py

        #        dsopts['file'] = str(dsopts['file']).format(n_iter=n_iter)
        h5_auxfile = h5io.WESTPAH5File(dsopts['file'].format(n_iter=n_iter))
        h5group = group
        if ("iter_" + str(n_iter).zfill(8)) not in h5_auxfile:
            h5_auxfile.create_group("iter_" + str(n_iter).zfill(8))
        group = h5_auxfile[('/' + "iter_" + str(n_iter).zfill(8))]

    h5path = dsopts['h5path']
    containing_group_name = posixpath.dirname(h5path)
    h5_dsname = posixpath.basename(h5path)

    # ensure arguments are sane
    if not shape and data is None:
        raise ValueError('either shape or data must be provided')
    elif data is None and (shape and dtype is None):
        raise ValueError('both shape and dtype must be provided when data is not provided')
    elif shape and data is not None and not data.shape == shape:
        raise ValueError('explicit shape {!r} does not match data shape {!r}'.format(shape, data.shape))

    if data is not None:
        shape = data.shape
        if dtype is None:
            dtype = data.dtype
    # end argument sanity checks

    # figure out where to store this data
    if containing_group_name:
        containing_group = group.require_group(containing_group_name)
    else:
        containing_group = group

    # has user requested an explicit data type?
    # the extra np.dtype is an idempotent operation on true dtype
    # objects, but ensures that things like np.float32, which are
    # actually NOT dtype objects, become dtype objects
    h5_dtype = np.dtype(dsopts.get('dtype', dtype))

    compression = None
    scaleoffset = None
    shuffle = False

    # compress if 1) explicitly requested, or 2) dataset size exceeds threshold and
    # compression not explicitly prohibited
    compression_directive = dsopts.get('compression')
    if compression_directive is None:
        # No directive
        nbytes = np.multiply.reduce(shape) * h5_dtype.itemsize
        if autocompress_threshold and nbytes > autocompress_threshold:
            compression = 9
    elif compression_directive == 0:  # includes False
        # Compression prohibited
        compression = None
    else:  # compression explicitly requested
        compression = compression_directive

    # Is scale/offset requested?
    scaleoffset = dsopts.get('scaleoffset', None)
    if scaleoffset is not None:
        scaleoffset = int(scaleoffset)

    # We always shuffle if we compress (losslessly)
    if compression:
        shuffle = True
    else:
        shuffle = False

    need_chunks = any([compression, scaleoffset is not None, shuffle])

    # We use user-provided chunks if available
    chunks_directive = dsopts.get('chunks')
    if chunks_directive is None:
        chunks = None
    elif chunks_directive is True:
        chunks = calc_chunksize(shape, h5_dtype)
    elif chunks_directive is False:
        chunks = None
    else:
        chunks = tuple(chunks_directive[i] if chunks_directive[i] <= shape[i] else shape[i] for i in range(len(shape)))

    if not chunks and need_chunks:
        chunks = calc_chunksize(shape, h5_dtype)

    opts = {'shape': shape, 'dtype': h5_dtype, 'compression': compression, 'shuffle': shuffle, 'chunks': chunks}

    try:
        import h5py._hl.filters

        h5py._hl.filters._COMP_FILTERS['scaleoffset']
    except (ImportError, KeyError, AttributeError):
        # filter not available, or an unexpected version of h5py
        # use lossless compression instead
        opts['compression'] = True
    else:
        opts['scaleoffset'] = scaleoffset

    if log.isEnabledFor(logging.DEBUG):
        log.debug('requiring aux dataset {!r}, shape={!r}, opts={!r}'.format(h5_dsname, shape, opts))

    dset = containing_group.require_dataset(h5_dsname, **opts)

    if data is not None:
        dset[...] = data

    if 'file' in list(dsopts.keys()):
        import h5py

        if dsopts['h5path'] not in h5group:
            h5group[dsopts['h5path']] = h5py.ExternalLink(
                dsopts['file'].format(n_iter=n_iter), ("/" + "iter_" + str(n_iter).zfill(8) + "/" + dsopts['h5path'])
            )

    return dset


def require_dataset_from_dsopts(group, dsopts, shape=None, dtype=None, data=None, autocompress_threshold=None, n_iter=None):
    if not dsopts.get('store', True):
        return None
    try:
        return group[dsopts['h5path']]
    except KeyError:
        return create_dataset_from_dsopts(
            group, dsopts, shape=shape, dtype=dtype, data=data, autocompress_threshold=autocompress_threshold, n_iter=n_iter
        )


def calc_chunksize(shape, dtype, max_chunksize=262144):
    '''Calculate a chunk size for HDF5 data, anticipating that access will slice
    along lower dimensions sooner than higher dimensions.'''

    chunk_shape = list(shape)
    for idim in range(len(shape)):
        chunk_nbytes = np.multiply.reduce(chunk_shape) * dtype.itemsize
        while chunk_shape[idim] > 1 and chunk_nbytes > max_chunksize:
            chunk_shape[idim] >>= 1  # divide by 2
            chunk_nbytes = np.multiply.reduce(chunk_shape) * dtype.itemsize

        if chunk_nbytes <= max_chunksize:
            break

    chunk_shape = tuple(chunk_shape)
    log.debug(
        'selected chunk shape {} for data set of type {} shaped {} (chunk size = {} bytes)'.format(
            chunk_shape, dtype, shape, chunk_nbytes
        )
    )
    return chunk_shape



import numpy

NAN = float('nan')


class RunningStatsAccumulator:
    def __init__(self, shape, dtype=numpy.float64, count_dtype=numpy.uint, weight_dtype=numpy.float64, mask_value=NAN):
        self.sum = numpy.zeros(shape, dtype)
        self.sqsum = numpy.zeros(shape, dtype)
        self.weight = numpy.zeros(shape, weight_dtype)
        self.count = numpy.zeros(shape, count_dtype)
        self.mask_value = mask_value

    def incorporate(self, index, value, weight):
        self.count[index] += 1
        self.weight[index] += weight
        self.sum[index] += weight * value
        self.sqsum[index] += weight * value * value

    def average(self):
        valid = self.count > 0
        avg = numpy.empty_like(self.sum)
        avg[valid] = self.sum[valid] / self.weight[valid]
        avg[~valid] = self.mask_value
        return avg

    mean = average

    def std(self):
        valid = self.count > 0
        vavg = self.average()[valid]
        std = numpy.empty_like(self.sqsum)
        std[valid] = (self.sqsum[valid] / self.weight[valid] - vavg * vavg) ** 0.5
        std[~valid] = self.mask_value
        return std



"""westext.stringmethod - Plugin to drive the adaptive evolution of one or more
strings of Voronoi bins

Joshua L. Adelman 2011
"""

from .string_method import DefaultStringMethod, WESTStringMethod
from .string_driver import StringDriver


__all__ = ['DefaultStringMethod', 'WESTStringMethod', 'StringDriver']



from westpa.tools import WESTMasterCommand, WESTParallelTool
from warnings import warn

from westpa.cli.tools.w_direct import DKinetics

# Just a shim to make sure everything works and is backwards compatible.


class WKinetics(DKinetics):
    subcommand = 'trace'
    help_text = 'averages and CIs for path-tracing kinetics analysis'
    default_output_file = 'kintrace.h5'


class WDirect(WESTMasterCommand, WESTParallelTool):
    prog = 'w_kinetics'
    subcommands = [WKinetics]
    subparsers_title = 'calculate state-to-state kinetics by tracing trajectories'
    description = '''\
Calculate state-to-state rates and transition event durations by tracing
trajectories.

A bin assignment file (usually "assign.h5") including trajectory labeling
is required (see "w_assign --help" for information on generating this file).

The output generated by this program is used as input for the ``w_kinavg``
tool, which converts the flux data in the output file into average rates
with confidence intervals. See ``w_kinavg trace --help`` for more
information.

-----------------------------------------------------------------------------
Output format
-----------------------------------------------------------------------------

The output file (-o/--output, by default "kintrace.h5") contains the
following datasets:

  ``/conditional_fluxes`` [iteration][state][state]
    *(Floating-point)* Macrostate-to-macrostate fluxes. These are **not**
    normalized by the population of the initial macrostate.

  ``/conditional_arrivals`` [iteration][stateA][stateB]
    *(Integer)* Number of trajectories arriving at state *stateB* in a given
    iteration, given that they departed from *stateA*.

  ``/total_fluxes`` [iteration][state]
    *(Floating-point)* Total flux into a given macrostate.

  ``/arrivals`` [iteration][state]
    *(Integer)* Number of trajectories arriving at a given state in a given
    iteration, regardless of where they originated.

  ``/duration_count`` [iteration]
    *(Integer)* The number of event durations recorded in each iteration.

  ``/durations`` [iteration][event duration]
    *(Structured -- see below)*  Event durations for transition events ending
    during a given iteration. These are stored as follows:

      istate
        *(Integer)* Initial state of transition event.
      fstate
        *(Integer)* Final state of transition event.
      duration
        *(Floating-point)* Duration of transition, in units of tau.
      weight
        *(Floating-point)* Weight of trajectory at end of transition, **not**
        normalized by initial state population.

Because state-to-state fluxes stored in this file are not normalized by
initial macrostate population, they cannot be used as rates without further
processing. The ``w_kinavg`` command is used to perform this normalization
while taking statistical fluctuation and correlation into account. See
``w_kinavg trace --help`` for more information.  Target fluxes (total flux
into a given state) require no such normalization.

-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------
'''


def entry_point():
    warn('{} is being deprecated.  Please use w_direct instead.'.format(WDirect.prog))
    import sys

    try:
        if sys.argv[1] != 'trace':
            sys.argv.insert(1, 'trace')
    except Exception:
        sys.argv.insert(1, 'trace')
    WDirect().main()


if __name__ == '__main__':
    entry_point()



import logging

from westpa.core.binning.binless import BinlessMapper
from westpa.core.sim_manager import WESimManager, grouper
from westpa.core.states import InitialState, pare_basis_initial_states
from westpa.core import wm_ops
from westpa.core.segment import Segment
import numpy as np

log = logging.getLogger(__name__)


class BinlessSimManager(WESimManager):
    def initialize_simulation(self, basis_states, target_states, start_states, segs_per_state=1, suppress_we=False):
        if len(target_states) > 0:
            if isinstance(self.system.bin_mapper, BinlessMapper):
                log.error("BinlessMapper cannot be an outer binning scheme with a target state\n")

        super().initialize_simulation(
            basis_states, target_states, start_states, segs_per_state=segs_per_state, suppress_we=suppress_we
        )

    def propagate(self):
        log.debug("BinlessManager in use")
        segments = list(self.incomplete_segments.values())
        log.debug('iteration {:d}: propagating {:d} segments'.format(self.n_iter, len(segments)))

        # all futures dispatched for this iteration
        futures = set()
        segment_futures = set()

        # Immediately dispatch any necessary initial state generation
        istate_gen_futures = self.get_istate_futures()
        futures.update(istate_gen_futures)

        # Dispatch propagation tasks using work manager
        for segment_block in grouper(self.propagator_block_size, segments):
            segment_block = [_f for _f in segment_block if _f]
            pbstates, pistates = pare_basis_initial_states(
                self.current_iter_bstates, list(self.current_iter_istates.values()), segment_block
            )
            future = self.work_manager.submit(wm_ops.propagate, args=(pbstates, pistates, segment_block))
            futures.add(future)
            segment_futures.add(future)

        while futures:
            # TODO: add capacity for timeout or SIGINT here
            future = self.work_manager.wait_any(futures)
            futures.remove(future)

            if future in segment_futures:
                segment_futures.remove(future)
                incoming = future.get_result()
                self.n_propagated += 1

                self.segments.update({segment.seg_id: segment for segment in incoming})
                self.completed_segments.update({segment.seg_id: segment for segment in incoming})

                new_istate_futures = self.get_istate_futures()
                istate_gen_futures.update(new_istate_futures)
                futures.update(new_istate_futures)

                with self.data_manager.expiring_flushing_lock():
                    self.data_manager.update_segments(self.n_iter, incoming)

            elif future in istate_gen_futures:
                istate_gen_futures.remove(future)
                _basis_state, initial_state = future.get_result()
                log.debug('received newly-prepared initial state {!r}'.format(initial_state))
                initial_state.istate_status = InitialState.ISTATE_STATUS_PREPARED
                with self.data_manager.expiring_flushing_lock():
                    self.data_manager.update_initial_states([initial_state], n_iter=self.n_iter + 1)
                self.we_driver.avail_initial_states[initial_state.state_id] = initial_state
            else:
                log.error('unknown future {!r} received from work manager'.format(future))
                raise AssertionError('untracked future {!r}'.format(future))

        # Collectively assign all segments to their bins...
        self.we_driver.assign(self.segments.values())

        # For cases where we need even more istates for recycled trajectories
        # futures should be empty at this point.
        istate_gen_futures = self.get_istate_futures()
        futures.update(istate_gen_futures)

        # Wait for istate_gen_futures and catch untracked futures.
        while futures:
            future = self.work_manager.wait_any(futures)
            futures.remove(future)

            if future in istate_gen_futures:
                istate_gen_futures.remove(future)
                _basis_state, initial_state = future.get_result()
                log.debug('received newly-prepared initial state {!r}'.format(initial_state))
                initial_state.istate_status = InitialState.ISTATE_STATUS_PREPARED
                with self.data_manager.expiring_flushing_lock():
                    self.data_manager.update_initial_states([initial_state], n_iter=self.n_iter + 1)
                self.we_driver.avail_initial_states[initial_state.state_id] = initial_state
            else:
                log.error('unknown future {!r} received from work manager'.format(future))
                raise AssertionError('untracked future {!r}'.format(future))

        log.debug('done with propagation')
        self.save_bin_data()
        self.data_manager.flush_backing()

    def prepare_iteration(self):
        log.debug('beginning iteration {:d}'.format(self.n_iter))

        # the WE driver needs a list of all target states for this iteration
        # along with information about any new weights introduced (e.g. by recycling)
        target_states = self.data_manager.get_target_states(self.n_iter)
        new_weights = self.data_manager.get_new_weight_data(self.n_iter)

        self.we_driver.new_iteration(target_states=target_states, new_weights=new_weights)

        # Get basis states used in this iteration
        self.current_iter_bstates = self.data_manager.get_basis_states(self.n_iter)

        # Get the segments for this iteration and separate into complete and incomplete
        if self.segments is None:
            segments = self.segments = {segment.seg_id: segment for segment in self.data_manager.get_segments()}
            log.debug('loaded {:d} segments'.format(len(segments)))
        else:
            segments = self.segments
            log.debug('using {:d} pre-existing segments'.format(len(segments)))

        completed_segments = self.completed_segments = {}
        incomplete_segments = self.incomplete_segments = {}
        for segment in segments.values():
            if segment.status == Segment.SEG_STATUS_COMPLETE:
                completed_segments[segment.seg_id] = segment
            else:
                incomplete_segments[segment.seg_id] = segment
        log.debug('{:d} segments are complete; {:d} are incomplete'.format(len(completed_segments), len(incomplete_segments)))

        if len(incomplete_segments) == len(segments):
            # Starting a new iteration
            self.rc.pstatus('Beginning iteration {:d}'.format(self.n_iter))
        elif incomplete_segments:
            self.rc.pstatus('Continuing iteration {:d}'.format(self.n_iter))
        self.rc.pstatus(
            '{:d} segments remain in iteration {:d} ({:d} total)'.format(len(incomplete_segments), self.n_iter, len(segments))
        )

        # Get the initial states active for this iteration (so that the propagator has them if necessary)
        self.current_iter_istates = {
            state.state_id: state for state in self.data_manager.get_segment_initial_states(list(segments.values()))
        }
        log.debug('This iteration uses {:d} initial states'.format(len(self.current_iter_istates)))

        # Assign this iteration's segments' initial points to bins and report on bin population
        initial_pcoords = self.system.new_pcoord_array(len(segments))
        initial_binning = self.system.bin_mapper.construct_bins()
        for iseg, segment in enumerate(segments.values()):
            initial_pcoords[iseg] = segment.pcoord[0]
        initial_assignments = self.system.bin_mapper.assign(initial_pcoords)
        for segment, assignment in zip(iter(segments.values()), initial_assignments):
            initial_binning[assignment].add(segment)
        self.report_bin_statistics(initial_binning, [], save_summary=True)
        del initial_pcoords, initial_binning

        self.rc.pstatus("Binless scheme in use")

        self.rc.pstatus("Waiting for segments to complete...")

        # Let the WE driver assign completed segments
        if completed_segments and len(incomplete_segments) == 0:
            self.we_driver.assign(list(completed_segments.values()))

        # load restart data
        self.data_manager.prepare_segment_restarts(
            incomplete_segments.values(), self.current_iter_bstates, self.current_iter_istates
        )

        # Get the basis states and initial states for the next iteration, necessary for doing on-the-fly recycling
        self.next_iter_bstates = self.data_manager.get_basis_states(self.n_iter + 1)
        self.next_iter_bstate_cprobs = np.add.accumulate([bstate.probability for bstate in self.next_iter_bstates])

        self.we_driver.avail_initial_states = {
            istate.state_id: istate for istate in self.data_manager.get_unused_initial_states(n_iter=self.n_iter + 1)
        }
        log.debug('{:d} unused initial states found'.format(len(self.we_driver.avail_initial_states)))

        # Invoke callbacks
        self.invoke_callbacks(self.prepare_iteration)

        # dispatch and immediately wait on result for prep_iter
        log.debug('dispatching propagator prep_iter to work manager')
        self.work_manager.submit(wm_ops.prep_iter, args=(self.n_iter, segments)).get_result()



import logging

import numpy as np

import westpa
from westpa.core import extloader
from westpa.core.yamlcfg import check_bool, ConfigItemMissing
from westpa.core.binning import VoronoiBinMapper

log = logging.getLogger(__name__)


class AdaptiveVoronoiDriver:
    '''
    This plugin implements an adaptive scheme using voronoi bins from
    Zhang 2010, J Chem Phys, 132. The options exposed to the configuration
    file are:

      - av_enabled (bool, default False): Enables adaptive binning
      - max_centers (int, default 10): The maximum number of voronoi centers to be placed
      - walk_count (integer, default 5): Number of walkers per voronoi center
      - center_freq (ingeter, default 1): Frequency of center placement
      - priority (integer, default 1): Priority in the plugin order
      - dfunc_method (function, non-optional, no default): Non-optional user defined
          function that will be used to calculate distances between voronoi centers and
          data points
      - mapper_func (function, optional): Optional user defined function for building bin
          mappers for more complicated binning schemes e.g. embedding the voronoi binning
          in a portion of the state space. If not defined the plugin will build a
          VoronoiBinMapper with the information it has.
    '''

    def __init__(self, sim_manager, plugin_config):
        if not sim_manager.work_manager.is_master:
            return

        self.sim_manager = sim_manager
        self.data_manager = sim_manager.data_manager
        self.system = sim_manager.system

        # Parameters from config file
        # this enables the adaptive voronoi, allows turning adaptive scheme off
        self.doAdaptiveVoronoi = check_bool(plugin_config.get('av_enabled', False))
        # sets maximim number of centers/voronoi bins
        self.max_centers = plugin_config.get('max_centers', 10)
        # sets number of walkers per bin/voronoi center
        self.walk_count = plugin_config.get('walk_count', 5)
        # center placement frequency in number of iterations
        self.center_freq = plugin_config.get('center_freq', 1)
        # priority of the plugin (allows for order of execution)
        self.priority = plugin_config.get('priority', 0)
        # pulls the distance function that will be used by the plugin
        self.dfunc = self.get_dfunc_method(plugin_config)
        # pulls a user defined function to build the next bin mapper
        self.mapper_func = self.get_mapper_func(plugin_config)

        # Get initial set of Voronoi centers
        self.centers = self.get_initial_centers()
        self.ncenters = len(self.centers)

        # Update the BinMapper
        self.update_bin_mapper()

        westpa.rc.pstatus('-adaptive voronoi mapping --------------\n')
        westpa.rc.pstatus('enabled: {}\n'.format(self.doAdaptiveVoronoi))
        westpa.rc.pstatus('max centers: {}\n'.format(self.max_centers))
        westpa.rc.pstatus('center adding freq: {}\n'.format(self.center_freq))
        westpa.rc.pstatus('centers: {}\n'.format(self.centers))
        westpa.rc.pstatus('----------------------------------------\n')
        westpa.rc.pflush()

        # Register callback
        if self.doAdaptiveVoronoi:
            sim_manager.register_callback(sim_manager.prepare_new_iteration, self.prepare_new_iteration, self.priority)

    def dfunc(self):
        '''
        Distance function to be used by the plugin. This function
        will be used to calculate the distance between each point.
        '''
        raise NotImplementedError

    def get_dfunc_method(self, plugin_config):
        try:
            methodname = plugin_config['dfunc_method']
        except KeyError:
            raise ConfigItemMissing('dfunc_method')

        dfunc_method = extloader.get_object(methodname)

        log.info('loaded adaptive voronoi dfunc method {!r}'.format(dfunc_method))

        return dfunc_method

    def get_mapper_func(self, plugin_config):
        try:
            methodname = plugin_config['mapper_func']
        except KeyError:
            return False

        mapper_func = extloader.get_object(methodname)

        log.info('loaded adaptive voronoi mapper function {!r}'.format(mapper_func))

        return mapper_func

    def get_initial_centers(self):
        '''
        This function pulls from the centers from either the
        previous bin mapper  or uses the definition from the
        system to calculate the number of centers
        '''
        self.data_manager.open_backing()

        with self.data_manager.lock:
            n_iter = max(self.data_manager.current_iteration - 1, 1)
            iter_group = self.data_manager.get_iter_group(n_iter)

            # First attempt to initialize voronoi centers
            # from data rather than system
            centers = None
            try:
                log.info('Voronoi centers from previous bin mapper')
                binhash = iter_group.attrs['binhash']
                bin_mapper = self.data_manager.get_bin_mapper(binhash)

                centers = bin_mapper.centers

            except Exception:
                log.warning(
                    'Initializing voronoi centers from data failed; \
                        Using definition in system instead.'
                )
                centers = self.system.bin_mapper.centers

        self.data_manager.close_backing()
        return centers

    def update_bin_mapper(self):
        '''Update the bin_mapper using the current set of voronoi centers'''

        westpa.rc.pstatus('westext.adaptvoronoi: Updating bin mapper\n')
        westpa.rc.pflush()

        # self.mapper_func = plugin_config.get('mapper_func', False)
        try:
            dfargs = getattr(self.system, 'dfargs', None)
            dfkwargs = getattr(self.system, 'dfkwargs', None)
            if self.mapper_func:
                # The mapper should take in 1) distance function,
                # 2) centers, 3) dfargs, 4) dfkwargs and return
                # the mapper we want
                self.system.bin_mapper = self.mapper_func(self.dfunc, self.centers, dfargs=dfargs, dfkwargs=dfkwargs)
            else:
                self.system.bin_mapper = VoronoiBinMapper(self.dfunc, self.centers, dfargs=dfargs, dfkwargs=dfkwargs)
            self.ncenters = self.system.bin_mapper.nbins
            new_target_counts = np.empty((self.ncenters,), int)
            new_target_counts[...] = self.walk_count
            self.system.bin_target_counts = new_target_counts
        except (ValueError, TypeError) as e:
            log.error(
                'AdaptiveVoronoiDriver Error: \
                    Failed updating the bin mapper: {}'.format(
                    e
                )
            )
            raise

    def update_centers(self, iter_group):
        '''
        Update the set of Voronoi centers according to
        Zhang 2010, J Chem Phys, 132. A short description
        of the algorithm can be found in the text:

        1) First reference structure is chosen randomly from
        the first set of given structure
        2) Given a set of n reference structures, for each
        configuration in the iteration the distances to each
        reference structure is calculated and the minimum
        distance is found
        3) The configuration with the minimum distance is
        selected as the next reference
        '''

        westpa.rc.pstatus('westext.adaptvoronoi: Updating Voronoi centers\n')
        westpa.rc.pflush()

        # Pull the current coordinates to find distances
        curr_pcoords = iter_group['pcoord']
        # Initialize distance array
        dists = np.zeros(curr_pcoords.shape[0])
        for iwalk, walk in enumerate(curr_pcoords):
            # Calculate distances using the provided function
            # and find the distance to the closest center
            dists[iwalk] = min(self.dfunc(walk[-1], self.centers))
        # Find the maximum of the minimum distances
        max_ind = np.where(dists == dists.max())
        # Use the maximum progress coordinate as our next center
        self.centers = np.vstack((self.centers, curr_pcoords[max_ind[0][0]][-1]))

    def prepare_new_iteration(self):
        n_iter = self.sim_manager.n_iter

        with self.data_manager.lock:
            iter_group = self.data_manager.get_iter_group(n_iter)

        # Check if we are at the correct frequency for updating the bin mapper
        if n_iter % self.center_freq == 0:
            # Check if we still need to add more centers
            if self.ncenters < self.max_centers:
                # First find the center to add
                self.update_centers(iter_group)
                # Update the bin mapper with the new center
                self.update_bin_mapper()



'''
Created on Jun 11, 2015

@author: mzwier
'''

import logging

log = logging.getLogger(__name__)

from .core import ZMQCore, Message, PassiveMultiTimer, IsNode

import zmq
from zmq.devices import ThreadProxy


class ZMQNode(ZMQCore, IsNode):
    def __init__(self, upstream_rr_endpoint, upstream_ann_endpoint, n_local_workers=None):
        ZMQCore.__init__(self)
        IsNode.__init__(self, n_local_workers)

        self.upstream_rr_endpoint = upstream_rr_endpoint
        self.upstream_ann_endpoint = upstream_ann_endpoint

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_traceback):
        return False

    def run(self):
        self.startup()

    @property
    def is_master(self):
        return False

    def comm_loop(self):
        self.context = zmq.Context.instance()
        # or else the proxies create sockets in a different context

        self.context.linger = 100
        # So we don't have to destroy the context at the end of the loop

        rr_proxy = ThreadProxy(zmq.ROUTER, zmq.DEALER)

        # We use push/pull so (1) we don't miss any announcements
        # and (2) we don't have to deal with subscription messages
        ann_proxy = ThreadProxy(zmq.SUB, zmq.PUB, zmq.PUSH)
        ann_monitor = self.context.socket(zmq.PULL)

        # Not monitoring request/reply streams for two reasons:
        # (1) we'd need to strip identity frames to interpret the messages
        # (2) interpreting the messages means we'd have to decode (unpickle) and then re-encode
        # all of the data flying through here, which seems like a waste just to see if
        # clients start up. We miss the edge failure case where one node's workers
        # start up but another's fail. Seems much less likely than all workers
        # failing to start up, which would be caught by the master

        ann_mon_endpoint = 'inproc://{:x}'.format(id(ann_monitor))
        ann_monitor.bind(ann_mon_endpoint)

        rr_proxy.bind_in(self.downstream_rr_endpoint)
        if self.local_rr_endpoint:
            rr_proxy.bind_in(self.local_rr_endpoint)
        self.log.debug('connecting upstream_rr_endpoint = {!r}'.format(self.upstream_rr_endpoint))
        rr_proxy.connect_out(self.upstream_rr_endpoint)

        ann_proxy.bind_out(self.downstream_ann_endpoint)
        if self.local_ann_endpoint:
            ann_proxy.bind_out(self.local_ann_endpoint)
        ann_proxy.connect_in(self.upstream_ann_endpoint)
        self.log.debug('connecting upstream_ann_endpoint = {!r}'.format(self.upstream_ann_endpoint))
        ann_proxy.setsockopt_in(zmq.SUBSCRIBE, b'')
        ann_proxy.connect_mon(ann_mon_endpoint)

        rr_proxy.start()
        ann_proxy.start()

        ann_monitor.connect(ann_mon_endpoint)

        inproc_socket = self.context.socket(zmq.SUB)
        inproc_socket.setsockopt(zmq.SUBSCRIBE, b'')
        inproc_socket.bind(self.inproc_endpoint)

        timers = PassiveMultiTimer()
        timers.add_timer('master_beacon', self.master_beacon_period)
        timers.add_timer('startup_timeout', self.startup_timeout)
        timers.reset()

        self.log.debug('master beacon period: {!r}'.format(self.master_beacon_period))
        self.log.debug('startup timeout: {!r}'.format(self.startup_timeout))

        peer_found = False

        poller = zmq.Poller()
        poller.register(ann_monitor, zmq.POLLIN)
        poller.register(inproc_socket, zmq.POLLIN)
        try:
            while True:
                poll_results = dict(poller.poll((timers.next_expiration_in() or 0.001) * 1000))

                if inproc_socket in poll_results:
                    msgs = self.recv_all(ann_monitor, validate=False)
                    if Message.SHUTDOWN in (msg.message for msg in msgs):
                        self.log.debug('shutdown received')
                        break

                if ann_monitor in poll_results:
                    msgs = self.recv_all(ann_monitor, validate=False)
                    message_tags = {msg.message for msg in msgs}
                    if Message.SHUTDOWN in message_tags:
                        self.log.debug('shutdown received')
                        break
                    if not peer_found and (Message.MASTER_BEACON in message_tags or Message.TASKS_AVAILABLE in message_tags):
                        peer_found = True
                        timers.remove_timer('startup_timeout')

                if not peer_found and timers.expired('startup_timeout'):
                    self.log.error('startup phase elapsed with no contact from peer; shutting down')
                    break

        finally:
            self.log.debug('exiting')
            self.context = None
            self.remove_ipc_endpoints()
            IsNode.shutdown(self)

    def startup(self):
        IsNode.startup(self)
        super().startup()



import numpy as np
from westpa.core.states import BasisState, TargetState, InitialState, return_state_type
from westpa.core.segment import Segment
from io import StringIO


class TestStates:
    def test_gen_bstate_from_file(self):
        bstate_txt = """
        # Comment
        0 0.1 a.rst
        1 1e-3 b.rst

        """
        true_answers = [['0', 1e-1, 'a.rst'], ['1', 1e-3, 'b.rst']]

        fake_file = StringIO(bstate_txt)
        bstates = BasisState.states_from_file(fake_file)

        for answer, state in zip(true_answers, bstates):
            assert state.label == answer[0]
            assert state.probability == answer[1]
            assert state.auxref == answer[2]

    def test_gen_tstate_from_file(self):
        tstate_txt = """
        # Comment
        0 1 1
        1 2 3

        """
        true_answers = [['0', np.array([1.0, 1.0])], ['1', np.array([2.0, 3.0])]]

        fake_file = StringIO(tstate_txt)
        tstates = TargetState.states_from_file(fake_file, float)

        for answer, state in zip(true_answers, tstates):
            assert state.label == answer[0]
            assert np.all(state.pcoord == answer[1])

    def test_return_state_type_bstate(self):
        bstate = BasisState(label=0, probability=0.1, state_id=10)
        state_name, state_id = return_state_type(bstate)

        assert state_name == 'BasisState'
        assert state_id == 10

    def test_return_state_type_istate(self):
        istate = InitialState(state_id=2, basis_state_id=10, iter_created=3)
        state_name, state_id = return_state_type(istate)

        assert state_name == 'InitialState'
        assert state_id == 2

    def test_return_state_type_segment(self):
        segment = Segment(seg_id=3)
        state_name, state_id = return_state_type(segment)

        assert state_name == 'Segment'
        assert state_id == 3

    def test_return_state_type_unknown(self):
        state_name, state_id = return_state_type(3)

        assert state_name == 'Unknown'
        assert state_id == float('inf')



'''
Tools for Monte Carlo bootstrap error analysis
'''

import logging
import math

import numpy as np
from numpy.random import Generator, MT19937

import westpa
from westpa.oldtools.aframe import AnalysisMixin

log = logging.getLogger(__name__)


def msort(input_array):
    return np.sort(input_array, axis=0)


class MCBSMixin(AnalysisMixin):
    def __init__(self):
        super().__init__()
        self.mcbs_alpha = None
        self.mcbs_nsets = None
        self.mcbs_display_confidence = None

    def add_args(self, parser, upcall=True):
        if upcall:
            try:
                upfunc = super().add_args
            except AttributeError:
                pass
            else:
                upfunc(parser)
        group = parser.add_argument_group('Monte Carlo bootstrap options')
        group.add_argument(
            '--confidence',
            dest='mcbs_confidence',
            type=float,
            default=0.95,
            metavar='P',
            help='''Construct a confidence interval of width P (default: 0.95=95%%).''',
        )
        group.add_argument(
            '--bssize',
            dest='mcbs_nsets',
            type=int,
            metavar='NSETS',
            help='''Use NSETS synthetic data sets to calculate confidence intervals (default:
                                   calculated based on confidence level, but not less than 1000).''',
        )

    def process_args(self, args, upcall=True):
        self.mcbs_alpha = 1 - args.mcbs_confidence
        self.mcbs_nsets = args.mcbs_size if args.mcbs_nsets else min(1000, calc_mcbs_nsets(self.mcbs_alpha))
        self.mcbs_display_confidence = '{:.{cp}f}'.format(
            100 * args.mcbs_confidence, cp=-int(math.floor(math.log10(self.mcbs_alpha))) - 2
        )
        westpa.rc.pstatus(
            'Using bootstrap of {:d} sets to calculate {:s}% confidence interval (alpha={:g}).'.format(
                self.mcbs_nsets, self.mcbs_display_confidence, self.mcbs_alpha
            )
        )

        if upcall:
            try:
                upfunc = super().process_args
            except AttributeError:
                pass
            else:
                upfunc(args)

    def calc_mcbs_nsets(self, alpha=None):
        alpha = alpha or self.mcbs_alpha
        return calc_mcbs_nsets(alpha)

    def calc_ci_bound_indices(self, n_sets=None, alpha=None):
        n_sets = n_sets or self.mcbs_nsets
        alpha = alpha or self.mcbs_alpha
        return calc_ci_bound_indices(n_sets, alpha)


ciinfo_dtype = np.dtype([('expectation', np.float64), ('ci_lower', np.float64), ('ci_upper', np.float64)])


def calc_mcbs_nsets(alpha):
    '''Return a bootstrap data set size appropriate for the given confidence level.'''
    return int(10 ** (math.ceil(-math.log10(alpha)) + 1))


def calc_ci_bound_indices(n_sets, alpha):
    return (int(math.floor(n_sets * alpha / 2)), int(math.ceil(n_sets * (1 - alpha / 2))))


def bootstrap_ci_ll(estimator, data, alpha, n_sets, storage, sort, eargs=(), ekwargs={}, fhat=None):
    '''Low-level routine for calculating bootstrap error estimates.  Arguments and return values are as those for
    ``bootstrap_ci``, except that no argument is optional except additional arguments for the estimator (``eargs``, ``ekwargs``).
    ``data`` must be an array (or subclass), and an additional array ``storage`` must be provided, which
    must be appropriately shaped and typed to hold ``n_sets`` results from ``estimator``.  Further, if the
    value ``fhat`` of the estimator must be pre-calculated to allocate ``storage``, then its value may be
    passed; otherwise, ``estimator(data,*eargs,**kwargs)`` will be called to calculate it.'''

    rng = Generator(MT19937())

    if fhat is None:
        fhat = estimator(data, *eargs, **ekwargs)
    dlen = len(data)

    for iset in range(n_sets):
        indices = rng.integers(dlen, size=(dlen,))
        storage[iset] = estimator(data[indices], *eargs, **ekwargs)

    synth_sorted = sort(storage)
    lbi = int(math.floor(n_sets * alpha / 2))
    ubi = int(math.ceil(n_sets * (1 - alpha / 2)))

    lb = synth_sorted[lbi]
    ub = synth_sorted[ubi]

    try:
        return (fhat, lb, ub, ub - lb, abs((ub - lb) / fhat) if fhat else 0, max(ub - fhat, fhat - lb))
    finally:
        del fhat, lb, ub, indices


def bootstrap_ci(estimator, data, alpha, n_sets=None, sort=msort, eargs=(), ekwargs={}):
    '''Perform a Monte Carlo bootstrap of a (1-alpha) confidence interval for the given ``estimator``.
    Returns (fhat, ci_lower, ci_upper), where fhat is the result of ``estimator(data, *eargs, **ekwargs)``,
    and ``ci_lower`` and ``ci_upper`` are the lower and upper bounds of the surrounding confidence
    interval, calculated by calling ``estimator(syndata, *eargs, **ekwargs)`` on each synthetic data
    set ``syndata``.  If ``n_sets`` is provided, that is the number of synthetic data sets generated,
    otherwise an appropriate size is selected automatically (see ``calc_mcbs_nsets()``).

    ``sort``, if given, is applied to sort the results of calling ``estimator`` on each
    synthetic data set prior to obtaining the confidence interval. This function must sort
    on the last index.

    Individual entries in synthetic data sets are selected by the first index of ``data``, allowing this
    function to be used on arrays of multidimensional data.

    Returns (fhat, lb, ub, ub-lb, abs((ub-lb)/fhat), and max(ub-fhat,fhat-lb)) (that is, the estimated value, the
    lower and upper bounds of the confidence interval, the width of the confidence interval, the relative
    width of the confidence interval, and the symmetrized error bar of the confidence interval).'''

    data = np.asanyarray(data)
    fhat = np.squeeze(estimator(data, *eargs, **ekwargs))
    n_sets = n_sets or calc_mcbs_nsets(alpha)
    fsynth = np.empty((n_sets,), dtype=fhat.dtype)
    try:
        return bootstrap_ci_ll(estimator, data, alpha, n_sets or calc_mcbs_nsets(alpha), fsynth, sort, eargs, ekwargs, fhat)
    finally:
        del fsynth



import logging
from typing import List, Optional
import numpy as np
import westpa
from westpa.core.binning import FuncBinMapper
from os.path import expandvars

log = logging.getLogger(__name__)


class MABBinMapper(FuncBinMapper):
    """
    Adaptively place bins between minimum and maximum segments along
    the progress coordinate. Extrema and bottleneck segments are assigned
    to their own bins.
    """

    def __init__(
        self,
        nbins: List[int],
        direction: Optional[List[int]] = None,
        skip: Optional[List[int]] = None,
        bottleneck: bool = True,
        pca: bool = False,
        mab_log: bool = False,
        bin_log: bool = False,
        bin_log_path: str = "$WEST_SIM_ROOT/binbounds.log",
    ):
        """
        Parameters
        ----------
        nbins : list of int
            List of number of bins in each dimension.
        direction : Optional[list of int], default: None
            List of directions in each dimension. Direction options:
                0   : default split at leading and lagging boundaries
                1   : split at leading boundary only
                -1  : split at lagging boundary only
                86  : no splitting at either leading or lagging boundary (both bottlenecks included)
        skip : Optional[list of int], default: None
            List of skip flags for each dimension. Default None (no skipping).
        bottleneck : bool, default: True
            Whether to enable bottleneck walker splitting.
        pca : bool, default: False
            Whether to perform PCA on progress coordinates before bin assignment.
        mab_log : bool, default: False
            Whether to output MAB info to west.log.
        bin_log : bool, default: False
            Whether to output MAB bin boundaries to a log file.
        bin_log_path : str, default: "$WEST_SIM_ROOT/binbounds.log"
            Path to output bin boundaries.
        """
        # Verifying parameters
        if nbins is None:
            raise ValueError("nbins is missing")
        ndim = len(nbins)

        direction = direction or [0] * ndim
        if len(direction) != ndim:
            direction = [0] * ndim
            log.warning("Direction list is not the correct dimensions, setting to defaults.")

        skip = skip or [0] * ndim
        if len(skip) != ndim:
            skip = [0] * ndim
            log.warning("Skip list is not the correct dimensions, setting to defaults.")

        kwargs = dict(
            nbins_per_dim=nbins,
            direction=direction,
            skip=skip,
            bottleneck=bottleneck,
            pca=pca,
            mab_log=mab_log,
            bin_log=bin_log,
            bin_log_path=bin_log_path,
        )

        n_total_bins = self.determine_total_bins(**kwargs)

        super().__init__(map_mab, n_total_bins, kwargs=kwargs)

    def determine_total_bins(
        self, nbins_per_dim: List[int], direction: List[int], skip: List[int], bottleneck: bool, **kwargs
    ) -> int:
        """
        Calculate the total number of bins needed, taking direction and skipping into account.
        This function is necessary because functional bin mappers need to "reserve"
        bins and tell the sim manager how many bins they will need to use, this is
        determined by taking all direction/skipping info into account.

        Parameters
        ----------
        nbins_per_dim : list of int
            Number of total bins in each dimension within the linear portion.
        direction : list of int
            Direction in each dimension.
        skip : list of int
            List indicating whether to skip each dimension.
        bottleneck : bool
            Whether to include a separate bin for bottleneck walker(s).
        **kwargs : dict
            Additional MAB parameters (unused).

        Returns
        -------
        n_total_bins : int
            Number of total bins.
        """
        # Update nbins_per_dim with any skipped dimensions, setting number of bins along skipped dimensions to 1
        skip = np.array([bool(s) for s in skip])
        nbins_per_dim = np.array(nbins_per_dim)
        nbins_per_dim[skip] = 1

        # Total bins is product of all linear bins plus and special bins
        n_total_bins = nbins_per_dim.prod()
        for direct, skip_dim in zip(direction, skip):
            if not skip_dim:
                if direct in [-1, 1]:
                    # 1 lead or lag bin + 1 bottleneck bin
                    n_total_bins += 1 + 1 * bottleneck
                elif direct == 0:
                    # 2 lead/lag bins + 2 bottleneck bins
                    n_total_bins += 2 + 2 * bottleneck
                elif direct == 86:
                    # 0 lead/lag + 2 bottleneck bins
                    n_total_bins += 2 * bottleneck
        return n_total_bins


def map_mab(coords: np.ndarray, mask: np.ndarray, output: List[int], *args, **kwargs) -> List[int]:
    """
    Adaptively place bins based on extrema and bottleneck segments along the progress coordinate.

    Bottleneck segments are where the difference in probability is the greatest
    along the progress coordinate. Operates per dimension (unless skipped) and places a fixed number of
    evenly spaced bins between the segments with the min and max pcoord values. Extrema and
    bottleneck segments are assigned their own bins.

    Parameters
    ----------
    coords : np.ndarray
        An array with pcoord and weight info.
    mask : np.ndarray
        Boolean array to filter out unwanted segments.
    output : list
        The main list that, for each segment, holds the bin assignment.
    *args : list
        Additional arguments.
    **kwargs : dict
        Additional keyword arguments. Contains most of the MAB-needed parameters.

    Returns
    ------
    output : list
        List with bin assignments for each segment.
    """

    # Argument Processing
    nbins_per_dim = kwargs.get("nbins_per_dim")
    ndim = len(nbins_per_dim)
    pca = kwargs.get("pca", False)
    bottleneck = kwargs.get("bottleneck", True)
    direction = kwargs.get("direction", [0] * ndim)
    skip = kwargs.get("skip", [0] * ndim)
    mab_log = kwargs.get("mab_log", False)
    bin_log = kwargs.get("bin_log", False)
    bin_log_path = kwargs.get("bin_log_path", "$WEST_SIM_ROOT/binbounds.log")

    if not np.any(mask):
        return output

    if skip is None:
        skip = [0] * ndim

    allcoords = coords.copy()
    allmask = mask.copy()

    weights = None
    isfinal = None
    splitting = False
    report = False

    # the segments should be sent in by the driver as half initial segments and half final segments
    # allcoords contains all segments
    # coords should contain ONLY final segments
    if coords.shape[1] > ndim:
        if coords[0, -1] == 0:
            report = True
        if coords.shape[1] > ndim + 1:
            isfinal = allcoords[:, ndim + 1].astype(bool)
        else:
            isfinal = np.ones(coords.shape[0], dtype=bool)
        coords = coords[isfinal, :ndim]
        weights = allcoords[isfinal, ndim]
        mask = mask[isfinal]
        splitting = True

    if not np.any(mask):
        coords = allcoords[:, :ndim]
        mask = allmask
        weights = None
        splitting = False

    originalcoords = np.copy(coords)
    if pca and len(output) > 1:
        coords = apply_pca(coords, weights)

    # Computing special bins (bottleneck and boundary bins)
    minlist, maxlist, bottlenecks_forward, bottlenecks_reverse = calculate_bin_boundaries(
        originalcoords, weights, mask, skip, splitting, bottleneck
    )

    if mab_log and report:
        log_mab_stats(minlist, maxlist, direction, skip)

    # Assign segments to bins
    n_bottleneck_filled = bin_assignment(
        allcoords,
        allmask,
        minlist,
        maxlist,
        bottlenecks_forward,
        bottlenecks_reverse,
        nbins_per_dim,
        direction,
        skip,
        splitting,
        bottleneck,
        output,
    )

    # Report MAB bin statistics
    if bin_log and report and westpa.rc.sim_manager.n_iter:
        log_bin_boundaries(
            skip,
            bottleneck,
            direction,
            bin_log_path,
            minlist,
            maxlist,
            nbins_per_dim,
            n_bottleneck_filled,
            bottlenecks_forward,
            bottlenecks_reverse,
        )

    return output


def apply_pca(coords, weights):
    colavg = np.mean(coords, axis=0)
    varcoords = coords - colavg
    covcoords = np.cov(varcoords.T, aweights=weights)
    eigval, eigvec = np.linalg.eigh(covcoords)
    eigvec = eigvec[:, np.argmax(np.abs(eigvec), axis=1)]
    eigvec[:, np.diag(eigvec) < 0] *= -1
    return np.dot(varcoords, eigvec)


def calculate_bin_boundaries(coords, weights, mask, skip, splitting, bottleneck):
    """
    This function calculates minima, maxima, and bottleneck segments.
    """
    skip = np.array([bool(s) for s in skip])

    # Initialize lists to hold minima and maxima along each dimension
    minlist, maxlist = [], []
    # Initialize lists to hold bottleneck segments along each dimension
    bottlenecks_forward, bottlenecks_reverse = [None] * len(coords[0]), [None] * len(coords[0])
    # number of unmasked coords
    n_coords = mask.sum()
    # Grabbing all unmasked coords and weights
    unmasked_coords = coords[mask, :]
    unmasked_weights = weights[mask] if weights is not None else None
    # Replace any zero weights with non-zero values so that log(weight) is well-defined
    if unmasked_weights is not None:
        unmasked_weights[unmasked_weights == 0] = 10**-323
    # Looping over each dimension of progress coordinate, even those being skipped
    for n in range(len(coords[0])):
        # We calculate the min and max pcoord along each dimension (boundary segments) even if skipping
        maxlist.append(np.max(coords[mask, n]))
        minlist.append(np.min(coords[mask, n]))
        # Now we calculate the bottleneck segments
        if splitting and bottleneck and not skip[n]:
            bottlenecks_forward[n], bottlenecks_reverse[n] = detect_bottlenecks(unmasked_coords, unmasked_weights, n_coords, n)

    return minlist, maxlist, bottlenecks_forward, bottlenecks_reverse


def detect_bottlenecks(unmasked_coords, unmasked_weights, n_coords, n):
    """
    Detect the bottleneck segments along the given coordinate n, this uses the weights
    """
    # Grabbing all unmasked coords in current dimension, plus corresponding weights
    # Sort by current dimension in coord, smallest to largest
    sorted_indices = unmasked_coords[:, n].argsort(kind='stable')

    # Grab sorted coords and weights
    coords_srt = unmasked_coords[sorted_indices, :]
    weights_srt = unmasked_weights[sorted_indices]

    # Also sort in reverse order for opposite direction
    coords_srt_flip = np.flipud(coords_srt)
    weights_srt_flip = np.flipud(weights_srt)

    # Initialize the max directional differences along current dimension as None (these may not be updated)
    bottleneck_coords, bottleneck_coords_flip = None, None
    maxdiff, maxdiff_flip = -np.inf, -np.inf

    # Looping through all non-boundary coords
    # Compute the cumulative weight on either side of each non-boundary walker
    for i in range(1, n_coords - 1):
        # Summing up weights of all walkers ahead of current walker along current dim in both directions
        cumulative_prob = np.sum(weights_srt[i + 1 :])
        cumulative_prob_flip = np.sum(weights_srt_flip[i + 1 :])
        # Compute the difference of log cumulative weight of current walker and all walkers ahead of it (Z im the MAB paper)
        # We use the log as weights vary over many orders of magnitude
        # Note a negative Z indicates the cumulative weight ahead of the current walker is larger than the weight of the current walker,
        # while a positive Z indicates the cumulative weight ahead of the current walker is smaller, indicating a barrier
        Z = np.log(weights_srt[i]) - np.log(cumulative_prob)
        Z_flip = np.log(weights_srt_flip[i]) - np.log(cumulative_prob_flip)
        # Update ALL coords of the current walker into bottlenecks_forward if it is largest
        # This way we uniquely identify a walker by its full set of coordinates
        if Z > maxdiff:
            bottleneck_coords = coords_srt[i, :]
            maxdiff = Z
        if Z_flip > maxdiff_flip:
            bottleneck_coords_flip = coords_srt_flip[i, :]
            maxdiff_flip = Z_flip
    return bottleneck_coords, bottleneck_coords_flip


def log_mab_stats(minlist, maxlist, direction, skip):
    westpa.rc.pstatus("################ MAB stats ################")
    westpa.rc.pstatus(f"minima in each dimension:      {minlist}")
    westpa.rc.pstatus(f"maxima in each dimension:      {maxlist}")
    westpa.rc.pstatus(f"direction in each dimension:   {direction}")
    westpa.rc.pstatus(f"skip in each dimension:        {skip}")
    westpa.rc.pstatus("###########################################")
    westpa.rc.pflush()


def bin_assignment(
    coords,
    mask,
    minlist,
    maxlist,
    bottlenecks_forward,
    bottlenecks_reverse,
    nbins_per_dim,
    direction,
    skip,
    splitting,
    bottleneck,
    output,
):
    """
    Assign segments to bins based on the minima, maxima, and
    bottleneck segments along the progress coordinate.
    """
    # Update nbins_per_dim with any skipped dimensions, setting number of bins along skipped dimensions to 1
    skip = np.array([bool(s) for s in skip])
    nbins_per_dim = np.array(nbins_per_dim)
    nbins_per_dim[skip] = 1
    direction = np.array(direction)

    ndim = len(nbins_per_dim)
    n_bottleneck_filled = 0

    # Boolean arrays that track use of special bins along each dimension
    skip_bneck_fwd = np.array([d == -1 if bottleneck else True for d in direction]) + skip
    skip_bneck_rev = np.array([d == 1 if bottleneck else True for d in direction]) + skip
    skip_lead = np.array([d in [86, -1] for d in direction]) + skip
    skip_lag = np.array([d in [86, 1] for d in direction]) + skip

    # List of dimensions that are not skipped
    active_dims = np.array([n for n in range(ndim) if not skip[n]])

    # Compute the boundary bin ID offsets
    # In forward direction, this is all the linear bins
    boundary_bin_id_offset_fwd = nbins_per_dim.prod()
    # In reverse, we add the number of forward boundary bins to the offset
    boundary_bin_id_offset_rev = boundary_bin_id_offset_fwd + (~skip_lead).sum()

    # Compute the bottleneck bin ID offsets
    # In forward direction, bin IDs are offset by all linear and boundary bins
    bneck_bin_id_offset_fwd = boundary_bin_id_offset_rev + (~skip_lag).sum()
    # In reverse, we add the number of forward bottleneck bins to the offset
    bneck_bin_id_offset_rev = bneck_bin_id_offset_fwd + (~skip_bneck_fwd).sum()

    # Bin assignment loop over all walkers
    for i in range(len(output)):
        # Skip masked walkers, these walkers bin IDs are unchanged
        if not mask[i]:
            continue
        # Initialize bin ID and special tracker for current coord
        # The special variable indicates a boundary or bottleneck walker (not assigned to the linear space)
        bin_id, special = 0, False

        # Searching for bottleneck bins first
        if splitting and bottleneck:
            for n in active_dims:
                # Grab coord(s) of current walker
                coord = coords[i][:ndim]
                # Assign bottlenecks, taking directionality into account
                # Check both directions when using 0 or 86
                # Note: 86 implies no leading or lagging bins, but does add bottlenecks for *both* directions when bottleneck is enabled
                # Note: All bottleneck bins will typically be filled unless a walker is simultaneously in bottleneck bins along multiple dimensions
                # or there are too few walkers to compute free energy barriers
                if (coord == bottlenecks_forward[n]).all() and not skip_bneck_fwd[n]:
                    bin_id = bneck_bin_id_offset_fwd + n - skip_bneck_fwd[:n].sum()
                    special = True
                    n_bottleneck_filled += 1
                    break
                elif (coord == bottlenecks_reverse[n]).all() and not skip_bneck_rev[n]:
                    bin_id = bneck_bin_id_offset_rev + n - skip_bneck_rev[:n].sum()
                    special = True
                    n_bottleneck_filled += 1
                    break

        # Now check for boundary walkers, taking directionality into account
        # This should only be done after fully checking for bottleneck walkers
        if splitting and not special:
            for n in active_dims:
                # Grab coord of current walker along current dimension
                coord = coords[i, n]
                if (coord == maxlist[n]) and not skip_lead[n]:
                    bin_id = boundary_bin_id_offset_fwd + n - skip_lead[:n].sum()
                    special = True
                    break
                elif (coord == minlist[n]) and not skip_lag[n]:
                    bin_id = boundary_bin_id_offset_rev + n - skip_lag[:n].sum()
                    special = True
                    break

        # Now check for linear bin walkers
        if not special:
            # Again we loop over the dimensions
            # Note: no need to worry about skipping as we've already set all skipped dimensions to 1 bin
            for n in range(ndim):
                coord = coords[i][n]
                nbins = nbins_per_dim[n]
                minp = minlist[n]
                maxp = maxlist[n]

                # Generate the bins along this dimension
                bins = np.linspace(minp, maxp, nbins + 1)

                # Assign walker to a bin along this dimension
                bin_number = np.digitize(coord, bins) - 1  # note np.digitize is 1-indexed

                # Sometimes the walker is exactly at the max/min value,
                # which would put it in the next bin
                if bin_number == nbins:
                    bin_number -= 1
                elif bin_number == -1:
                    bin_number = 0
                elif bin_number > nbins or bin_number < -1:
                    raise ValueError("Walker out of boundary.")

                # Assign to bin within the full dimensional space
                bin_id += bin_number * np.prod(nbins_per_dim[:n])

        # Output is the main list that, for each segment, holds the bin assignment
        output[i] = bin_id
    return n_bottleneck_filled


def log_bin_boundaries(
    skip,
    bottleneck,
    direction,
    bin_log_path,
    minlist,
    maxlist,
    nbins_per_dim,
    n_bottleneck_filled,
    bottlenecks_forward,
    bottlenecks_reverse,
):
    ndim = len(nbins_per_dim)
    skip = np.array([bool(s) for s in skip])
    active_dims = np.array([n for n in range(ndim) if not skip[n]])
    max_bottleneck = np.sum([1 if direction[n] in [-1, 1] else 2 for n in active_dims]) if bottleneck else 0
    with open(expandvars(bin_log_path), 'a') as bb_file:
        # Iteration Number
        bb_file.write(f'Iteration: {westpa.rc.sim_manager.n_iter}\n')
        bb_file.write('MAB linear bin boundaries: ')
        for n in range(ndim):
            # Write binbounds per dim
            bb_file.write(f'{np.linspace(minlist[n], maxlist[n], nbins_per_dim[n] + 1)}\t')
        # Min/Max pcoord
        bb_file.write(f'\nLagging pcoord in each dimension: {minlist}\n')
        bb_file.write(f'Leading pcoord in each dimension: {maxlist}\n')
        # Bottlenecks bins exist
        if bottleneck:
            bb_file.write(f'Number of bottleneck bins filled: {n_bottleneck_filled} / {max_bottleneck}\n')
            for n in active_dims:
                if direction[n] in [0, 1, 86]:
                    bb_file.write(f'Dimension {n} forward bottleneck walker at: {list(bottlenecks_forward[n])}\n')
                if direction[n] in [0, -1, 86]:
                    bb_file.write(f'Dimension {n} backward bottleneck walker at: {list(bottlenecks_reverse[n])}\n')
            bb_file.write('\n')
        else:
            bb_file.write('\n')



import sys

from setuptools import setup, Extension, find_packages

import versioneer


def extensions():
    from Cython.Build import cythonize
    from numpy import get_include as np_get_include

    np_inc = np_get_include()

    common_cflags = [
        '-O3',
    ]

    fasthist_module = Extension(
        'westpa.fasthist._fasthist',
        sources=['src/westpa/fasthist/_fasthist.pyx'],
        include_dirs=[np_inc],
        extra_compile_args=common_cflags,
    )

    trajtree_module = Extension(
        'westpa.trajtree._trajtree',
        sources=['src/westpa/trajtree/_trajtree.pyx'],
        include_dirs=[np_inc],
        extra_compile_args=common_cflags,
    )

    mclib_module = Extension(
        'westpa.mclib._mclib', sources=['src/westpa/mclib/_mclib.pyx'], include_dirs=[np_inc], extra_compile_args=common_cflags
    )

    binning_module = Extension(
        'westpa.core.binning._assign',
        sources=['src/westpa/core/binning/_assign.pyx'],
        include_dirs=[np_inc],
        extra_compile_args=common_cflags,
    )

    kinetics_module = Extension(
        'westpa.core.kinetics._kinetics',
        sources=['src/westpa/core/kinetics/_kinetics.pyx'],
        include_dirs=[np_inc],
        extra_compile_args=common_cflags,
    )

    reweight_module = Extension(
        'westpa.core.reweight._reweight',
        sources=['src/westpa/core/reweight/_reweight.pyx'],
        include_dirs=[np_inc],
        extra_compile_args=common_cflags,
    )

    exts = [fasthist_module, trajtree_module, mclib_module, binning_module, kinetics_module, reweight_module]

    exts = cythonize(exts, language_level=sys.version_info[0])

    return exts


console_scripts_core = [
    'w_fork = westpa.cli.core.w_fork:entry_point',
    'w_states = westpa.cli.core.w_states:entry_point',
    'w_run = westpa.cli.core.w_run:entry_point',
    'w_truncate = westpa.cli.core.w_truncate:entry_point',
    'w_init = westpa.cli.core.w_init:entry_point',
    'w_succ = westpa.cli.core.w_succ:entry_point',
]

console_scripts_tools = [
    'w_direct = westpa.cli.tools.w_direct:entry_point',
    'w_dumpsegs = westpa.cli.tools.w_dumpsegs:entry_point',
    'w_postanalysis_matrix = westpa.cli.tools.w_postanalysis_matrix:entry_point',
    'w_select = westpa.cli.tools.w_select:entry_point',
    'w_ntop = westpa.cli.tools.w_ntop:entry_point',
    'w_kinavg = westpa.cli.tools.w_kinavg:entry_point',
    'w_eddist = westpa.cli.tools.w_eddist:entry_point',
    'w_assign = westpa.cli.tools.w_assign:entry_point',
    'w_trace = westpa.cli.tools.w_trace:entry_point',
    'w_crawl = westpa.cli.tools.w_crawl:entry_point',
    'w_kinetics = westpa.cli.tools.w_kinetics:entry_point',
    'w_fluxanl = westpa.cli.tools.w_fluxanl:entry_point',
    'w_reweight = westpa.cli.tools.w_reweight:entry_point',
    'w_pdist = westpa.cli.tools.w_pdist:entry_point',
    'w_ipa = westpa.cli.tools.w_ipa:entry_point',
    'w_bins = westpa.cli.tools.w_bins:entry_point',
    'w_stateprobs = westpa.cli.tools.w_stateprobs:entry_point',
    'w_postanalysis_reweight = westpa.cli.tools.w_postanalysis_reweight:entry_point',
    'ploterr = westpa.cli.tools.ploterr:entry_point',
    'plothist = westpa.cli.tools.plothist:entry_point',
    'w_multi_west = westpa.cli.tools.w_multi_west:entry_point',
    'w_red = westpa.cli.tools.w_red:entry_point',
]

console_scripts = console_scripts_core + console_scripts_tools

CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: MIT License",
    "Operating System :: POSIX",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Cython",
]

INSTALL_REQUIRES = [
    "numpy >= 1.25.0, <3; python_version >='3.10'",
    "numpy >= 1.25.0, <2; python_version <'3.10'",  # pytables is broken for numpy2/py3.9
    "scipy >= 0.19.1",
    "h5py >= 2.10",
    "mdtraj >= 1.9.5",
    "pyyaml",
    "pyzmq",
    "matplotlib",
    "blessings",
    "ipykernel",
    "tqdm",
    "pandas",
    "tables",
]

EXTRAS_REQUIRE = {
    "tests": ["pytest", "pytest-cov", "pytest-rerunfailures", "pytest-timeout"],
    "mpi": ["mpi4py"],
}

EXTRAS_REQUIRE["dev"] = EXTRAS_REQUIRE["tests"] + ["pre-commit"]


metadata = dict(
    name='westpa',
    url='http://github.com/westpa/westpa',
    license='MIT',
    long_description=open('README.rst', encoding='utf8').read(),
    version=versioneer.get_version(),
    keywords='',
    cmdclass=versioneer.get_cmdclass(),
    python_requires=">=3.9",
    zip_safe=False,
    classifiers=CLASSIFIERS,
    entry_points={'console_scripts': console_scripts},
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    package_data={},
    packages=find_packages(where='src'),
    package_dir={"": "src"},
    description='WESTPA is a package for constructing and running stochastic simulations using the "weighted ensemble" approach of Huber and Kim (1996).',
)


if __name__ == '__main__':
    metadata['ext_modules'] = extensions()
    setup(**metadata)



import logging
import sys

from .core import WorkManager, WMFuture


log = logging.getLogger(__name__)


class SerialWorkManager(WorkManager):
    @classmethod
    def from_environ(cls, wmenv=None):
        return cls()

    def __init__(self):
        log.debug('initializing serial work manager')
        super().__init__()
        self.n_workers = 1

    def submit(self, fn, args=None, kwargs=None):
        ft = WMFuture()
        try:
            result = fn(*(args if args is not None else ()), **(kwargs if kwargs is not None else {}))
        except Exception as e:
            ft._set_exception(e, sys.exc_info()[2])
        else:
            ft._set_result(result)
        return ft



from westpa.tools import WESTMasterCommand, WESTParallelTool
from warnings import warn

from westpa.cli.tools.w_reweight import RWAverage

# Just a shim to make sure everything works and is backwards compatible.
# We're making sure it has the appropriate functions so that it can be called
# as a regular tool, and not a subcommand.


class PAAverage(RWAverage):
    subcommand = 'average'
    help_text = ''
    default_output_file = 'kinrw.h5'
    # This isn't strictly necessary, but for the moment, here it is.
    # We really need to modify the underlying class so that we don't pull this sort of stuff if it isn't necessary.
    # That'll take some case handling, which is fine.
    default_kinetics_file = 'flux_matrices.h5'


class WReweight(WESTMasterCommand, WESTParallelTool):
    prog = 'w_postanalysis_reweight'
    subcommands = [PAAverage]
    subparsers_title = 'calculate state-to-state kinetics by tracing trajectories'
    description = '''\
A convenience function to run kinetics/probs. Bin assignments,
including macrostate definitions, are required. (See
"w_assign --help" for more information).

For more information on the individual subcommands this subs in for, run
w_reweight {kinetics/probs} --help.

-----------------------------------------------------------------------------
Command-line options
-----------------------------------------------------------------------------
'''


def entry_point():
    warn('{} is being deprecated.  Please use w_reweight instead.'.format(WReweight.prog))
    # If we're not really supporting subcommands...
    import sys

    try:
        if sys.argv[1] != 'average':
            sys.argv.insert(1, 'average')
    except Exception:
        sys.argv.insert(1, 'average')
    WReweight().main()


if __name__ == '__main__':
    entry_point()



from time import time

import numpy as np
from numpy.random import Generator, MT19937

from westpa.fasthist import histnd


def test_double(npts=1024 * 1024, loops=3):
    rng = Generator(MT19937())  # RNG for this function

    mine_times = [None] * loops
    theirs_times = [None] * loops

    # though 1.0 should be sufficient, weirdness in the boundary conditions
    # for np.digitize appears to introduce a discrepancy, so throw something
    # greater than 1.0 in for good measure
    binbounds = [[0, 0.5, 1, 1.1] for _ in range(3)]
    weights = rng.random(size=npts)
    # weights = np.ones((npts,), dtype=np.float64)
    for n in range(loops):
        testdat = rng.random(size=(npts, 3))
        mstart = time()
        mine = histnd(testdat, binbounds, weights=weights)
        mstop = time()
        tstart = time()
        theirs = np.histogramdd(testdat, binbounds, weights=weights)[0]
        tstop = time()
        mine_times[n] = mstop - mstart
        theirs_times[n] = tstop - tstart
        print(mine)
        print(theirs)
        errsum = np.abs(mine - theirs).sum()
        errsum_per_item = errsum / npts
        rel_err = errsum / np.abs(weights).sum()
        print('sum of the absolute errors: {} ({} relative, {} per entry)'.format(errsum, rel_err, errsum_per_item))

    print('mine, best of {}:   {}'.format(loops, min(mine_times)))
    print('theirs, best of {}: {}'.format(loops, min(theirs_times)))

    assert min(mine_times) < min(theirs_times)


def test_float(npts=1024 * 1024, ndim=3, loops=3):
    rng = Generator(MT19937())  # RNG for this function

    mine_times = [None] * loops
    theirs_times = [None] * loops

    binbounds = [[0, 0.5, 1, 1.1] for _ in range(ndim)]
    # weights = rng.random(size=npts)
    weights = np.ones((npts,), dtype=np.float64)
    for n in range(loops):
        testdat = rng.random(size=(npts, ndim), dtype=np.float32)
        print(testdat)
        mstart = time()
        mine = histnd(testdat, binbounds, weights=weights)
        mstop = time()
        tstart = time()
        theirs = np.histogramdd(testdat, binbounds, weights=weights)[0]
        tstop = time()
        mine_times[n] = mstop - mstart
        theirs_times[n] = tstop - tstart
        print(mine)
        print(theirs)
        errsum = np.abs(mine - theirs).sum()
        errsum_per_item = errsum / npts
        rel_err = errsum / np.abs(weights).sum()
        print('sum of the absolute errors: {} ({} relative, {} per entry)'.format(errsum, rel_err, errsum_per_item))

    print('mine, best of {}:   {}'.format(loops, min(mine_times)))
    print('theirs, best of {}: {}'.format(loops, min(theirs_times)))

    assert min(mine_times) < min(theirs_times)


def test_uint(npts=1024 * 1024, ndim=3, loops=3):
    rng = Generator(MT19937())  # RNG for this function

    mine_times = [None] * loops
    theirs_times = [None] * loops

    binbounds = [[0, 1, 2, 3, 4] for x in range(ndim)]
    # weights = rng.random(size=npts)
    weights = np.ones((npts,), dtype=np.float64)
    print('binbounds: {}'.format(binbounds))
    print('weights')
    print(weights)
    for n in range(loops):
        testdat = rng.integers(0, 4, size=(npts, ndim), dtype=np.uint16)
        print('test data')
        print(testdat)
        mstart = time()
        mine = histnd(testdat, binbounds, weights=weights)
        mstop = time()
        tstart = time()
        theirs = np.histogramdd(testdat, binbounds, weights=weights)[0]
        tstop = time()
        mine_times[n] = mstop - mstart
        theirs_times[n] = tstop - tstart
        print(mine)
        print(theirs)
        errsum = np.abs(mine - theirs).sum()
        errsum_per_item = errsum / npts
        rel_err = errsum / np.abs(weights).sum()
        print('sum of the absolute errors: {} ({} relative, {} per entry)'.format(errsum, rel_err, errsum_per_item))

    print('mine, best of {}:   {}'.format(loops, min(mine_times)))
    print('theirs, best of {}: {}'.format(loops, min(theirs_times)))

    assert min(mine_times) < min(theirs_times)
